# mypy: allow-untyped-defs
from __future__ import annotations

import abc
from collections import defaultdict
from collections import deque
from collections import OrderedDict
from collections.abc import Callable
from collections.abc import Generator
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Mapping
from collections.abc import MutableMapping
from collections.abc import Sequence
from collections.abc import Set as AbstractSet
import dataclasses
import functools
import inspect
import os
from pathlib import Path
import sys
import types
from typing import Any
from typing import cast
from typing import Final
from typing import final
from typing import Generic
from typing import NoReturn
from typing import Optional
from typing import overload
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union
import warnings

import _pytest
from _pytest import nodes
from _pytest._code import getfslineno
from _pytest._code import Source
from _pytest._code.code import FormattedExcinfo
from _pytest._code.code import TerminalRepr
from _pytest._io import TerminalWriter
from _pytest.compat import assert_never
from _pytest.compat import get_real_func
from _pytest.compat import getfuncargnames
from _pytest.compat import getimfunc
from _pytest.compat import getlocation
from _pytest.compat import NOTSET
from _pytest.compat import NotSetType
from _pytest.compat import safe_getattr
from _pytest.compat import safe_isclass
from _pytest.config import _PluggyPlugin
from _pytest.config import Config
from _pytest.config import ExitCode
from _pytest.config.argparsing import Parser
from _pytest.deprecated import check_ispytest
from _pytest.deprecated import MARKED_FIXTURE
from _pytest.deprecated import YIELD_FIXTURE
from _pytest.main import Session
from _pytest.mark import Mark
from _pytest.mark import ParameterSet
from _pytest.mark.structures import MarkDecorator
from _pytest.outcomes import fail
from _pytest.outcomes import skip
from _pytest.outcomes import TEST_OUTCOME
from _pytest.pathlib import absolutepath
from _pytest.pathlib import bestrelpath
from _pytest.scope import _ScopeName
from _pytest.scope import HIGH_SCOPES
from _pytest.scope import Scope
from _pytest.warning_types import PytestRemovedIn9Warning
from _pytest.warning_types import PytestWarning


if sys.version_info < (3, 11):
    from exceptiongroup import BaseExceptionGroup


if TYPE_CHECKING:
    from _pytest.python import CallSpec2
    from _pytest.python import Function
    from _pytest.python import Metafunc


# The value of the fixture -- return/yield of the fixture function (type variable).
FixtureValue = TypeVar("FixtureValue")
# The type of the fixture function (type variable).
FixtureFunction = TypeVar("FixtureFunction", bound=Callable[..., object])
# The type of a fixture function (type alias generic in fixture value).
_FixtureFunc = Union[
    Callable[..., FixtureValue], Callable[..., Generator[FixtureValue]]
]
# The type of FixtureDef.cached_result (type alias generic in fixture value).
_FixtureCachedResult = Union[
    tuple[
        # The result.
        FixtureValue,
        # Cache key.
        object,
        None,
    ],
    tuple[
        None,
        # Cache key.
        object,
        # The exception and the original traceback.
        tuple[BaseException, Optional[types.TracebackType]],
    ],
]


@dataclasses.dataclass(frozen=True)
class PseudoFixtureDef(Generic[FixtureValue]):
    cached_result: _FixtureCachedResult[FixtureValue]
    _scope: Scope


def pytest_sessionstart(session: Session) -> None:
    session._fixturemanager = FixtureManager(session)


def get_scope_package(
    node: nodes.Item,
    fixturedef: FixtureDef[object],
) -> nodes.Node | None:
    from _pytest.python import Package

    for parent in node.iter_parents():
        if isinstance(parent, Package) and parent.nodeid == fixturedef.baseid:
            return parent
    return node.session


def get_scope_node(node: nodes.Node, scope: Scope) -> nodes.Node | None:
    """Get the closest parent node (including self) which matches the given
    scope.

    If there is no parent node for the scope (e.g. asking for class scope on a
    Module, or on a Function when not defined in a class), returns None.
    """
    import _pytest.python

    if scope is Scope.Function:
        # Type ignored because this is actually safe, see:
        # https://github.com/python/mypy/issues/4717
        return node.getparent(nodes.Item)  # type: ignore[type-abstract]
    elif scope is Scope.Class:
        return node.getparent(_pytest.python.Class)
    elif scope is Scope.Module:
        return node.getparent(_pytest.python.Module)
    elif scope is Scope.Package:
        return node.getparent(_pytest.python.Package)
    elif scope is Scope.Session:
        return node.getparent(_pytest.main.Session)
    else:
        assert_never(scope)


# TODO: Try to use FixtureFunctionDefinition instead of the marker
def getfixturemarker(obj: object) -> FixtureFunctionMarker | None:
    """Return fixturemarker or None if it doesn't exist"""
    if isinstance(obj, FixtureFunctionDefinition):
        return obj._fixture_function_marker
    return None


# Algorithm for sorting on a per-parametrized resource setup basis.
# It is called for Session scope first and performs sorting
# down to the lower scopes such as to minimize number of "high scope"
# setups and teardowns.


@dataclasses.dataclass(frozen=True)
class ParamArgKey:
    """A key for a high-scoped parameter used by an item.

    For use as a hashable key in `reorder_items`. The combination of fields
    is meant to uniquely identify a particular "instance" of a param,
    potentially shared by multiple items in a scope.
    """

    #: The param name.
    argname: str
    param_index: int
    #: For scopes Package, Module, Class, the path to the file (directory in
    #: Package's case) of the package/module/class where the item is defined.
    scoped_item_path: Path | None
    #: For Class scope, the class where the item is defined.
    item_cls: type | None


_V = TypeVar("_V")
OrderedSet = dict[_V, None]


def get_param_argkeys(item: nodes.Item, scope: Scope) -> Iterator[ParamArgKey]:
    """Return all ParamArgKeys for item matching the specified high scope."""
    assert scope is not Scope.Function

    try:
        callspec: CallSpec2 = item.callspec  # type: ignore[attr-defined]
    except AttributeError:
        return

    item_cls = None
    if scope is Scope.Session:
        scoped_item_path = None
    elif scope is Scope.Package:
        # Package key = module's directory.
        scoped_item_path = item.path.parent
    elif scope is Scope.Module:
        scoped_item_path = item.path
    elif scope is Scope.Class:
        scoped_item_path = item.path
        item_cls = item.cls  # type: ignore[attr-defined]
    else:
        assert_never(scope)

    for argname in callspec.indices:
        if callspec._arg2scope[argname] != scope:
            continue
        param_index = callspec.indices[argname]
        yield ParamArgKey(argname, param_index, scoped_item_path, item_cls)


def reorder_items(items: Sequence[nodes.Item]) -> list[nodes.Item]:
    argkeys_by_item: dict[Scope, dict[nodes.Item, OrderedSet[ParamArgKey]]] = {}
    items_by_argkey: dict[Scope, dict[ParamArgKey, OrderedDict[nodes.Item, None]]] = {}
    for scope in HIGH_SCOPES:
        scoped_argkeys_by_item = argkeys_by_item[scope] = {}
        scoped_items_by_argkey = items_by_argkey[scope] = defaultdict(OrderedDict)
        for item in items:
            argkeys = dict.fromkeys(get_param_argkeys(item, scope))
            if argkeys:
                scoped_argkeys_by_item[item] = argkeys
                for argkey in argkeys:
                    scoped_items_by_argkey[argkey][item] = None

    items_set = dict.fromkeys(items)
    return list(
        reorder_items_atscope(
            items_set, argkeys_by_item, items_by_argkey, Scope.Session
        )
    )


def reorder_items_atscope(
    items: OrderedSet[nodes.Item],
    argkeys_by_item: Mapping[Scope, Mapping[nodes.Item, OrderedSet[ParamArgKey]]],
    items_by_argkey: Mapping[
        Scope, Mapping[ParamArgKey, OrderedDict[nodes.Item, None]]
    ],
    scope: Scope,
) -> OrderedSet[nodes.Item]:
    if scope is Scope.Function or len(items) < 3:
        return items

    scoped_items_by_argkey = items_by_argkey[scope]
    scoped_argkeys_by_item = argkeys_by_item[scope]

    ignore: set[ParamArgKey] = set()
    items_deque = deque(items)
    items_done: OrderedSet[nodes.Item] = {}
    while items_deque:
        no_argkey_items: OrderedSet[nodes.Item] = {}
        slicing_argkey = None
        while items_deque:
            item = items_deque.popleft()
            if item in items_done or item in no_argkey_items:
                continue
            argkeys = dict.fromkeys(
                k for k in scoped_argkeys_by_item.get(item, ()) if k not in ignore
            )
            if not argkeys:
                no_argkey_items[item] = None
            else:
                slicing_argkey, _ = argkeys.popitem()
                # We don't have to remove relevant items from later in the
                # deque because they'll just be ignored.
                matching_items = [
                    i for i in scoped_items_by_argkey[slicing_argkey] if i in items
                ]
                for i in reversed(matching_items):
                    items_deque.appendleft(i)
                    # Fix items_by_argkey order.
                    for other_scope in HIGH_SCOPES:
                        other_scoped_items_by_argkey = items_by_argkey[other_scope]
                        for argkey in argkeys_by_item[other_scope].get(i, ()):
                            argkey_dict = other_scoped_items_by_argkey[argkey]
                            if not hasattr(sys, "pypy_version_info"):
                                argkey_dict[i] = None
                                argkey_dict.move_to_end(i, last=False)
                            else:
                                # Work around a bug in PyPy:
                                # https://github.com/pypy/pypy/issues/5257
                                # https://github.com/pytest-dev/pytest/issues/13312
                                bkp = argkey_dict.copy()
                                argkey_dict.clear()
                                argkey_dict[i] = None
                                argkey_dict.update(bkp)
                break
        if no_argkey_items:
            reordered_no_argkey_items = reorder_items_atscope(
                no_argkey_items, argkeys_by_item, items_by_argkey, scope.next_lower()
            )
            items_done.update(reordered_no_argkey_items)
        if slicing_argkey is not None:
            ignore.add(slicing_argkey)
    return items_done


@dataclasses.dataclass(frozen=True)
class FuncFixtureInfo:
    """Fixture-related information for a fixture-requesting item (e.g. test
    function).

    This is used to examine the fixtures which an item requests statically
    (known during collection). This includes autouse fixtures, fixtures
    requested by the `usefixtures` marker, fixtures requested in the function
    parameters, and the transitive closure of these.

    An item may also request fixtures dynamically (using `request.getfixturevalue`);
    these are not reflected here.
    """

    __slots__ = ("argnames", "initialnames", "name2fixturedefs", "names_closure")

    # Fixture names that the item requests directly by function parameters.
    argnames: tuple[str, ...]
    # Fixture names that the item immediately requires. These include
    # argnames + fixture names specified via usefixtures and via autouse=True in
    # fixture definitions.
    initialnames: tuple[str, ...]
    # The transitive closure of the fixture names that the item requires.
    # Note: can't include dynamic dependencies (`request.getfixturevalue` calls).
    names_closure: list[str]
    # A map from a fixture name in the transitive closure to the FixtureDefs
    # matching the name which are applicable to this function.
    # There may be multiple overriding fixtures with the same name. The
    # sequence is ordered from furthest to closes to the function.
    name2fixturedefs: dict[str, Sequence[FixtureDef[Any]]]

    def prune_dependency_tree(self) -> None:
        """Recompute names_closure from initialnames and name2fixturedefs.

        Can only reduce names_closure, which means that the new closure will
        always be a subset of the old one. The order is preserved.

        This method is needed because direct parametrization may shadow some
        of the fixtures that were included in the originally built dependency
        tree. In this way the dependency tree can get pruned, and the closure
        of argnames may get reduced.
        """
        closure: set[str] = set()
        working_set = set(self.initialnames)
        while working_set:
            argname = working_set.pop()
            # Argname may be something not included in the original names_closure,
            # in which case we ignore it. This currently happens with pseudo
            # FixtureDefs which wrap 'get_direct_param_fixture_func(request)'.
            # So they introduce the new dependency 'request' which might have
            # been missing in the original tree (closure).
            if argname not in closure and argname in self.names_closure:
                closure.add(argname)
                if argname in self.name2fixturedefs:
                    working_set.update(self.name2fixturedefs[argname][-1].argnames)

        self.names_closure[:] = sorted(closure, key=self.names_closure.index)


class FixtureRequest(abc.ABC):
    """The type of the ``request`` fixture.

    A request object gives access to the requesting test context and has a
    ``param`` attribute in case the fixture is parametrized.
    """

    def __init__(
        self,
        pyfuncitem: Function,
        fixturename: str | None,
        arg2fixturedefs: dict[str, Sequence[FixtureDef[Any]]],
        fixture_defs: dict[str, FixtureDef[Any]],
        *,
        _ispytest: bool = False,
    ) -> None:
        check_ispytest(_ispytest)
        #: Fixture for which this request is being performed.
        self.fixturename: Final = fixturename
        self._pyfuncitem: Final = pyfuncitem
        # The FixtureDefs for each fixture name requested by this item.
        # Starts from the statically-known fixturedefs resolved during
        # collection. Dynamically requested fixtures (using
        # `request.getfixturevalue("foo")`) are added dynamically.
        self._arg2fixturedefs: Final = arg2fixturedefs
        # The evaluated argnames so far, mapping to the FixtureDef they resolved
        # to.
        self._fixture_defs: Final = fixture_defs
        # Notes on the type of `param`:
        # -`request.param` is only defined in parametrized fixtures, and will raise
        #   AttributeError otherwise. Python typing has no notion of "undefined", so
        #   this cannot be reflected in the type.
        # - Technically `param` is only (possibly) defined on SubRequest, not
        #   FixtureRequest, but the typing of that is still in flux so this cheats.
        # - In the future we might consider using a generic for the param type, but
        #   for now just using Any.
        self.param: Any

    @property
    def _fixturemanager(self) -> FixtureManager:
        return self._pyfuncitem.session._fixturemanager

    @property
    @abc.abstractmethod
    def _scope(self) -> Scope:
        raise NotImplementedError()

    @property
    def scope(self) -> _ScopeName:
        """Scope string, one of "function", "class", "module", "package", "session"."""
        return self._scope.value

    @abc.abstractmethod
    def _check_scope(
        self,
        requested_fixturedef: FixtureDef[object] | PseudoFixtureDef[object],
        requested_scope: Scope,
    ) -> None:
        raise NotImplementedError()

    @property
    def fixturenames(self) -> list[str]:
        """Names of all active fixtures in this request."""
        result = list(self._pyfuncitem.fixturenames)
        result.extend(set(self._fixture_defs).difference(result))
        return result

    @property
    @abc.abstractmethod
    def node(self):
        """Underlying collection node (depends on current request scope)."""
        raise NotImplementedError()

    @property
    def config(self) -> Config:
        """The pytest config object associated with this request."""
        return self._pyfuncitem.config

    @property
    def function(self):
        """Test function object if the request has a per-function scope."""
        if self.scope != "function":
            raise AttributeError(
                f"function not available in {self.scope}-scoped context"
            )
        return self._pyfuncitem.obj

    @property
    def cls(self):
        """Class (can be None) where the test function was collected."""
        if self.scope not in ("class", "function"):
            raise AttributeError(f"cls not available in {self.scope}-scoped context")
        clscol = self._pyfuncitem.getparent(_pytest.python.Class)
        if clscol:
            return clscol.obj

    @property
    def instance(self):
        """Instance (can be None) on which test function was collected."""
        if self.scope != "function":
            return None
        return getattr(self._pyfuncitem, "instance", None)

    @property
    def module(self):
        """Python module object where the test function was collected."""
        if self.scope not in ("function", "class", "module"):
            raise AttributeError(f"module not available in {self.scope}-scoped context")
        mod = self._pyfuncitem.getparent(_pytest.python.Module)
        assert mod is not None
        return mod.obj

    @property
    def path(self) -> Path:
        """Path where the test function was collected."""
        if self.scope not in ("function", "class", "module", "package"):
            raise AttributeError(f"path not available in {self.scope}-scoped context")
        return self._pyfuncitem.path

    @property
    def keywords(self) -> MutableMapping[str, Any]:
        """Keywords/markers dictionary for the underlying node."""
        node: nodes.Node = self.node
        return node.keywords

    @property
    def session(self) -> Session:
        """Pytest session object."""
        return self._pyfuncitem.session

    @abc.abstractmethod
    def addfinalizer(self, finalizer: Callable[[], object]) -> None:
        """Add finalizer/teardown function to be called without arguments after
        the last test within the requesting test context finished execution."""
        raise NotImplementedError()

    def applymarker(self, marker: str | MarkDecorator) -> None:
        """Apply a marker to a single test function invocation.

        This method is useful if you don't want to have a keyword/marker
        on all function invocations.

        :param marker:
            An object created by a call to ``pytest.mark.NAME(...)``.
        """
        self.node.add_marker(marker)

    def raiseerror(self, msg: str | None) -> NoReturn:
        """Raise a FixtureLookupError exception.

        :param msg:
            An optional custom error message.
        """
        raise FixtureLookupError(None, self, msg)

    def getfixturevalue(self, argname: str) -> Any:
        """Dynamically run a named fixture function.

        Declaring fixtures via function argument is recommended where possible.
        But if you can only decide whether to use another fixture at test
        setup time, you may use this function to retrieve it inside a fixture
        or test function body.

        This method can be used during the test setup phase or the test run
        phase, but during the test teardown phase a fixture's value may not
        be available.

        :param argname:
            The fixture name.
        :raises pytest.FixtureLookupError:
            If the given fixture could not be found.
        """
        # Note that in addition to the use case described in the docstring,
        # getfixturevalue() is also called by pytest itself during item and fixture
        # setup to evaluate the fixtures that are requested statically
        # (using function parameters, autouse, etc).

        fixturedef = self._get_active_fixturedef(argname)
        assert fixturedef.cached_result is not None, (
            f'The fixture value for "{argname}" is not available.  '
            "This can happen when the fixture has already been torn down."
        )
        return fixturedef.cached_result[0]

    def _iter_chain(self) -> Iterator[SubRequest]:
        """Yield all SubRequests in the chain, from self up.

        Note: does *not* yield the TopRequest.
        """
        current = self
        while isinstance(current, SubRequest):
            yield current
            current = current._parent_request

    def _get_active_fixturedef(
        self, argname: str
    ) -> FixtureDef[object] | PseudoFixtureDef[object]:
        if argname == "request":
            cached_result = (self, [0], None)
            return PseudoFixtureDef(cached_result, Scope.Function)

        # If we already finished computing a fixture by this name in this item,
        # return it.
        fixturedef = self._fixture_defs.get(argname)
        if fixturedef is not None:
            self._check_scope(fixturedef, fixturedef._scope)
            return fixturedef

        # Find the appropriate fixturedef.
        fixturedefs = self._arg2fixturedefs.get(argname, None)
        if fixturedefs is None:
            # We arrive here because of a dynamic call to
            # getfixturevalue(argname) which was naturally
            # not known at parsing/collection time.
            fixturedefs = self._fixturemanager.getfixturedefs(argname, self._pyfuncitem)
            if fixturedefs is not None:
                self._arg2fixturedefs[argname] = fixturedefs
        # No fixtures defined with this name.
        if fixturedefs is None:
            raise FixtureLookupError(argname, self)
        # The are no fixtures with this name applicable for the function.
        if not fixturedefs:
            raise FixtureLookupError(argname, self)

        # A fixture may override another fixture with the same name, e.g. a
        # fixture in a module can override a fixture in a conftest, a fixture in
        # a class can override a fixture in the module, and so on.
        # An overriding fixture can request its own name (possibly indirectly);
        # in this case it gets the value of the fixture it overrides, one level
        # up.
        # Check how many `argname`s deep we are, and take the next one.
        # `fixturedefs` is sorted from furthest to closest, so use negative
        # indexing to go in reverse.
        index = -1
        for request in self._iter_chain():
            if request.fixturename == argname:
                index -= 1
        # If already consumed all of the available levels, fail.
        if -index > len(fixturedefs):
            raise FixtureLookupError(argname, self)
        fixturedef = fixturedefs[index]

        # Prepare a SubRequest object for calling the fixture.
        try:
            callspec = self._pyfuncitem.callspec
        except AttributeError:
            callspec = None
        if callspec is not None and argname in callspec.params:
            param = callspec.params[argname]
            param_index = callspec.indices[argname]
            # The parametrize invocation scope overrides the fixture's scope.
            scope = callspec._arg2scope[argname]
        else:
            param = NOTSET
            param_index = 0
            scope = fixturedef._scope
            self._check_fixturedef_without_param(fixturedef)
        # The parametrize invocation scope only controls caching behavior while
        # allowing wider-scoped fixtures to keep depending on the parametrized
        # fixture. Scope control is enforced for parametrized fixtures
        # by recreating the whole fixture tree on parameter change.
        # Hence `fixturedef._scope`, not `scope`.
        self._check_scope(fixturedef, fixturedef._scope)
        subrequest = SubRequest(
            self, scope, param, param_index, fixturedef, _ispytest=True
        )

        # Make sure the fixture value is cached, running it if it isn't
        fixturedef.execute(request=subrequest)

        self._fixture_defs[argname] = fixturedef
        return fixturedef

    def _check_fixturedef_without_param(self, fixturedef: FixtureDef[object]) -> None:
        """Check that this request is allowed to execute this fixturedef without
        a param."""
        funcitem = self._pyfuncitem
        has_params = fixturedef.params is not None
        fixtures_not_supported = getattr(funcitem, "nofuncargs", False)
        if has_params and fixtures_not_supported:
            msg = (
                f"{funcitem.name} does not support fixtures, maybe unittest.TestCase subclass?\n"
                f"Node id: {funcitem.nodeid}\n"
                f"Function type: {type(funcitem).__name__}"
            )
            fail(msg, pytrace=False)
        if has_params:
            frame = inspect.stack()[3]
            frameinfo = inspect.getframeinfo(frame[0])
            source_path = absolutepath(frameinfo.filename)
            source_lineno = frameinfo.lineno
            try:
                source_path_str = str(source_path.relative_to(funcitem.config.rootpath))
            except ValueError:
                source_path_str = str(source_path)
            location = getlocation(fixturedef.func, funcitem.config.rootpath)
            msg = (
                "The requested fixture has no parameter defined for test:\n"
                f"    {funcitem.nodeid}\n\n"
                f"Requested fixture '{fixturedef.argname}' defined in:\n"
                f"{location}\n\n"
                f"Requested here:\n"
                f"{source_path_str}:{source_lineno}"
            )
            fail(msg, pytrace=False)

    def _get_fixturestack(self) -> list[FixtureDef[Any]]:
        values = [request._fixturedef for request in self._iter_chain()]
        values.reverse()
        return values


@final
class TopRequest(FixtureRequest):
    """The type of the ``request`` fixture in a test function."""

    def __init__(self, pyfuncitem: Function, *, _ispytest: bool = False) -> None:
        super().__init__(
            fixturename=None,
            pyfuncitem=pyfuncitem,
            arg2fixturedefs=pyfuncitem._fixtureinfo.name2fixturedefs.copy(),
            fixture_defs={},
            _ispytest=_ispytest,
        )

    @property
    def _scope(self) -> Scope:
        return Scope.Function

    def _check_scope(
        self,
        requested_fixturedef: FixtureDef[object] | PseudoFixtureDef[object],
        requested_scope: Scope,
    ) -> None:
        # TopRequest always has function scope so always valid.
        pass

    @property
    def node(self):
        return self._pyfuncitem

    def __repr__(self) -> str:
        return f"<FixtureRequest for {self.node!r}>"

    def _fillfixtures(self) -> None:
        item = self._pyfuncitem
        for argname in item.fixturenames:
            if argname not in item.funcargs:
                item.funcargs[argname] = self.getfixturevalue(argname)

    def addfinalizer(self, finalizer: Callable[[], object]) -> None:
        self.node.addfinalizer(finalizer)


@final
class SubRequest(FixtureRequest):
    """The type of the ``request`` fixture in a fixture function requested
    (transitively) by a test function."""

    def __init__(
        self,
        request: FixtureRequest,
        scope: Scope,
        param: Any,
        param_index: int,
        fixturedef: FixtureDef[object],
        *,
        _ispytest: bool = False,
    ) -> None:
        super().__init__(
            pyfuncitem=request._pyfuncitem,
            fixturename=fixturedef.argname,
            fixture_defs=request._fixture_defs,
            arg2fixturedefs=request._arg2fixturedefs,
            _ispytest=_ispytest,
        )
        self._parent_request: Final[FixtureRequest] = request
        self._scope_field: Final = scope
        self._fixturedef: Final[FixtureDef[object]] = fixturedef
        if param is not NOTSET:
            self.param = param
        self.param_index: Final = param_index

    def __repr__(self) -> str:
        return f"<SubRequest {self.fixturename!r} for {self._pyfuncitem!r}>"

    @property
    def _scope(self) -> Scope:
        return self._scope_field

    @property
    def node(self):
        scope = self._scope
        if scope is Scope.Function:
            # This might also be a non-function Item despite its attribute name.
            node: nodes.Node | None = self._pyfuncitem
        elif scope is Scope.Package:
            node = get_scope_package(self._pyfuncitem, self._fixturedef)
        else:
            node = get_scope_node(self._pyfuncitem, scope)
        if node is None and scope is Scope.Class:
            # Fallback to function item itself.
            node = self._pyfuncitem
        assert node, (
            f'Could not obtain a node for scope "{scope}" for function {self._pyfuncitem!r}'
        )
        return node

    def _check_scope(
        self,
        requested_fixturedef: FixtureDef[object] | PseudoFixtureDef[object],
        requested_scope: Scope,
    ) -> None:
        if isinstance(requested_fixturedef, PseudoFixtureDef):
            return
        if self._scope > requested_scope:
            # Try to report something helpful.
            argname = requested_fixturedef.argname
            fixture_stack = "\n".join(
                self._format_fixturedef_line(fixturedef)
                for fixturedef in self._get_fixturestack()
            )
            requested_fixture = self._format_fixturedef_line(requested_fixturedef)
            fail(
                f"ScopeMismatch: You tried to access the {requested_scope.value} scoped "
                f"fixture {argname} with a {self._scope.value} scoped request object. "
                f"Requesting fixture stack:\n{fixture_stack}\n"
                f"Requested fixture:\n{requested_fixture}",
                pytrace=False,
            )

    def _format_fixturedef_line(self, fixturedef: FixtureDef[object]) -> str:
        factory = fixturedef.func
        path, lineno = getfslineno(factory)
        if isinstance(path, Path):
            path = bestrelpath(self._pyfuncitem.session.path, path)
        signature = inspect.signature(factory)
        return f"{path}:{lineno + 1}:  def {factory.__name__}{signature}"

    def addfinalizer(self, finalizer: Callable[[], object]) -> None:
        self._fixturedef.addfinalizer(finalizer)


@final
class FixtureLookupError(LookupError):
    """Could not return a requested fixture (missing or invalid)."""

    def __init__(
        self, argname: str | None, request: FixtureRequest, msg: str | None = None
    ) -> None:
        self.argname = argname
        self.request = request
        self.fixturestack = request._get_fixturestack()
        self.msg = msg

    def formatrepr(self) -> FixtureLookupErrorRepr:
        tblines: list[str] = []
        addline = tblines.append
        stack = [self.request._pyfuncitem.obj]
        stack.extend(map(lambda x: x.func, self.fixturestack))
        msg = self.msg
        # This function currently makes an assumption that a non-None msg means we
        # have a non-empty `self.fixturestack`. This is currently true, but if
        # somebody at some point want to extend the use of FixtureLookupError to
        # new cases it might break.
        # Add the assert to make it clearer to developer that this will fail, otherwise
        # it crashes because `fspath` does not get set due to `stack` being empty.
        assert self.msg is None or self.fixturestack, (
            "formatrepr assumptions broken, rewrite it to handle it"
        )
        if msg is not None:
            # The last fixture raise an error, let's present
            # it at the requesting side.
            stack = stack[:-1]
        for function in stack:
            fspath, lineno = getfslineno(function)
            try:
                lines, _ = inspect.getsourcelines(get_real_func(function))
            except (OSError, IndexError, TypeError):
                error_msg = "file %s, line %s: source code not available"
                addline(error_msg % (fspath, lineno + 1))
            else:
                addline(f"file {fspath}, line {lineno + 1}")
                for i, line in enumerate(lines):
                    line = line.rstrip()
                    addline("  " + line)
                    if line.lstrip().startswith("def"):
                        break

        if msg is None:
            fm = self.request._fixturemanager
            available = set()
            parent = self.request._pyfuncitem.parent
            assert parent is not None
            for name, fixturedefs in fm._arg2fixturedefs.items():
                faclist = list(fm._matchfactories(fixturedefs, parent))
                if faclist:
                    available.add(name)
            if self.argname in available:
                msg = (
                    f" recursive dependency involving fixture '{self.argname}' detected"
                )
            else:
                msg = f"fixture '{self.argname}' not found"
            msg += "\n available fixtures: {}".format(", ".join(sorted(available)))
            msg += "\n use 'pytest --fixtures [testpath]' for help on them."

        return FixtureLookupErrorRepr(fspath, lineno, tblines, msg, self.argname)


class FixtureLookupErrorRepr(TerminalRepr):
    def __init__(
        self,
        filename: str | os.PathLike[str],
        firstlineno: int,
        tblines: Sequence[str],
        errorstring: str,
        argname: str | None,
    ) -> None:
        self.tblines = tblines
        self.errorstring = errorstring
        self.filename = filename
        self.firstlineno = firstlineno
        self.argname = argname

    def toterminal(self, tw: TerminalWriter) -> None:
        # tw.line("FixtureLookupError: %s" %(self.argname), red=True)
        for tbline in self.tblines:
            tw.line(tbline.rstrip())
        lines = self.errorstring.split("\n")
        if lines:
            tw.line(
                f"{FormattedExcinfo.fail_marker}       {lines[0].strip()}",
                red=True,
            )
            for line in lines[1:]:
                tw.line(
                    f"{FormattedExcinfo.flow_marker}       {line.strip()}",
                    red=True,
                )
        tw.line()
        tw.line(f"{os.fspath(self.filename)}:{self.firstlineno + 1}")


def call_fixture_func(
    fixturefunc: _FixtureFunc[FixtureValue], request: FixtureRequest, kwargs
) -> FixtureValue:
    if inspect.isgeneratorfunction(fixturefunc):
        fixturefunc = cast(Callable[..., Generator[FixtureValue]], fixturefunc)
        generator = fixturefunc(**kwargs)
        try:
            fixture_result = next(generator)
        except StopIteration:
            raise ValueError(f"{request.fixturename} did not yield a value") from None
        finalizer = functools.partial(_teardown_yield_fixture, fixturefunc, generator)
        request.addfinalizer(finalizer)
    else:
        fixturefunc = cast(Callable[..., FixtureValue], fixturefunc)
        fixture_result = fixturefunc(**kwargs)
    return fixture_result


def _teardown_yield_fixture(fixturefunc, it) -> None:
    """Execute the teardown of a fixture function by advancing the iterator
    after the yield and ensure the iteration ends (if not it means there is
    more than one yield in the function)."""
    try:
        next(it)
    except StopIteration:
        pass
    else:
        fs, lineno = getfslineno(fixturefunc)
        fail(
            f"fixture function has more than one 'yield':\n\n"
            f"{Source(fixturefunc).indent()}\n"
            f"{fs}:{lineno + 1}",
            pytrace=False,
        )


def _eval_scope_callable(
    scope_callable: Callable[[str, Config], _ScopeName],
    fixture_name: str,
    config: Config,
) -> _ScopeName:
    try:
        # Type ignored because there is no typing mechanism to specify
        # keyword arguments, currently.
        result = scope_callable(fixture_name=fixture_name, config=config)  # type: ignore[call-arg]
    except Exception as e:
        raise TypeError(
            f"Error evaluating {scope_callable} while defining fixture '{fixture_name}'.\n"
            "Expected a function with the signature (*, fixture_name, config)"
        ) from e
    if not isinstance(result, str):
        fail(
            f"Expected {scope_callable} to return a 'str' while defining fixture '{fixture_name}', but it returned:\n"
            f"{result!r}",
            pytrace=False,
        )
    return result


@final
class FixtureDef(Generic[FixtureValue]):
    """A container for a fixture definition.

    Note: At this time, only explicitly documented fields and methods are
    considered public stable API.
    """

    def __init__(
        self,
        config: Config,
        baseid: str | None,
        argname: str,
        func: _FixtureFunc[FixtureValue],
        scope: Scope | _ScopeName | Callable[[str, Config], _ScopeName] | None,
        params: Sequence[object] | None,
        ids: tuple[object | None, ...] | Callable[[Any], object | None] | None = None,
        *,
        _ispytest: bool = False,
        # only used in a deprecationwarning msg, can be removed in pytest9
        _autouse: bool = False,
    ) -> None:
        check_ispytest(_ispytest)
        # The "base" node ID for the fixture.
        #
        # This is a node ID prefix. A fixture is only available to a node (e.g.
        # a `Function` item) if the fixture's baseid is a nodeid of a parent of
        # node.
        #
        # For a fixture found in a Collector's object (e.g. a `Module`s module,
        # a `Class`'s class), the baseid is the Collector's nodeid.
        #
        # For a fixture found in a conftest plugin, the baseid is the conftest's
        # directory path relative to the rootdir.
        #
        # For other plugins, the baseid is the empty string (always matches).
        self.baseid: Final = baseid or ""
        # Whether the fixture was found from a node or a conftest in the
        # collection tree. Will be false for fixtures defined in non-conftest
        # plugins.
        self.has_location: Final = baseid is not None
        # The fixture factory function.
        self.func: Final = func
        # The name by which the fixture may be requested.
        self.argname: Final = argname
        if scope is None:
            scope = Scope.Function
        elif callable(scope):
            scope = _eval_scope_callable(scope, argname, config)
        if isinstance(scope, str):
            scope = Scope.from_user(
                scope, descr=f"Fixture '{func.__name__}'", where=baseid
            )
        self._scope: Final = scope
        # If the fixture is directly parametrized, the parameter values.
        self.params: Final = params
        # If the fixture is directly parametrized, a tuple of explicit IDs to
        # assign to the parameter values, or a callable to generate an ID given
        # a parameter value.
        self.ids: Final = ids
        # The names requested by the fixtures.
        self.argnames: Final = getfuncargnames(func, name=argname)
        # If the fixture was executed, the current value of the fixture.
        # Can change if the fixture is executed with different parameters.
        self.cached_result: _FixtureCachedResult[FixtureValue] | None = None
        self._finalizers: Final[list[Callable[[], object]]] = []

        # only used to emit a deprecationwarning, can be removed in pytest9
        self._autouse = _autouse

    @property
    def scope(self) -> _ScopeName:
        """Scope string, one of "function", "class", "module", "package", "session"."""
        return self._scope.value

    def addfinalizer(self, finalizer: Callable[[], object]) -> None:
        self._finalizers.append(finalizer)

    def finish(self, request: SubRequest) -> None:
        exceptions: list[BaseException] = []
        while self._finalizers:
            fin = self._finalizers.pop()
            try:
                fin()
            except BaseException as e:
                exceptions.append(e)
        node = request.node
        node.ihook.pytest_fixture_post_finalizer(fixturedef=self, request=request)
        # Even if finalization fails, we invalidate the cached fixture
        # value and remove all finalizers because they may be bound methods
        # which will keep instances alive.
        self.cached_result = None
        self._finalizers.clear()
        if len(exceptions) == 1:
            raise exceptions[0]
        elif len(exceptions) > 1:
            msg = f'errors while tearing down fixture "{self.argname}" of {node}'
            raise BaseExceptionGroup(msg, exceptions[::-1])

    def execute(self, request: SubRequest) -> FixtureValue:
        """Return the value of this fixture, executing it if not cached."""
        # Ensure that the dependent fixtures requested by this fixture are loaded.
        # This needs to be done before checking if we have a cached value, since
        # if a dependent fixture has their cache invalidated, e.g. due to
        # parametrization, they finalize themselves and fixtures depending on it
        # (which will likely include this fixture) setting `self.cached_result = None`.
        # See #4871
        requested_fixtures_that_should_finalize_us = []
        for argname in self.argnames:
            fixturedef = request._get_active_fixturedef(argname)
            # Saves requested fixtures in a list so we later can add our finalizer
            # to them, ensuring that if a requested fixture gets torn down we get torn
            # down first. This is generally handled by SetupState, but still currently
            # needed when this fixture is not parametrized but depends on a parametrized
            # fixture.
            if not isinstance(fixturedef, PseudoFixtureDef):
                requested_fixtures_that_should_finalize_us.append(fixturedef)

        # Check for (and return) cached value/exception.
        if self.cached_result is not None:
            request_cache_key = self.cache_key(request)
            cache_key = self.cached_result[1]
            try:
                # Attempt to make a normal == check: this might fail for objects
                # which do not implement the standard comparison (like numpy arrays -- #6497).
                cache_hit = bool(request_cache_key == cache_key)
            except (ValueError, RuntimeError):
                # If the comparison raises, use 'is' as fallback.
                cache_hit = request_cache_key is cache_key

            if cache_hit:
                if self.cached_result[2] is not None:
                    exc, exc_tb = self.cached_result[2]
                    raise exc.with_traceback(exc_tb)
                else:
                    result = self.cached_result[0]
                    return result
            # We have a previous but differently parametrized fixture instance
            # so we need to tear it down before creating a new one.
            self.finish(request)
            assert self.cached_result is None

        # Add finalizer to requested fixtures we saved previously.
        # We make sure to do this after checking for cached value to avoid
        # adding our finalizer multiple times. (#12135)
        finalizer = functools.partial(self.finish, request=request)
        for parent_fixture in requested_fixtures_that_should_finalize_us:
            parent_fixture.addfinalizer(finalizer)

        ihook = request.node.ihook
        try:
            # Setup the fixture, run the code in it, and cache the value
            # in self.cached_result
            result = ihook.pytest_fixture_setup(fixturedef=self, request=request)
        finally:
            # schedule our finalizer, even if the setup failed
            request.node.addfinalizer(finalizer)

        return result

    def cache_key(self, request: SubRequest) -> object:
        return getattr(request, "param", None)

    def __repr__(self) -> str:
        return f"<FixtureDef argname={self.argname!r} scope={self.scope!r} baseid={self.baseid!r}>"


def resolve_fixture_function(
    fixturedef: FixtureDef[FixtureValue], request: FixtureRequest
) -> _FixtureFunc[FixtureValue]:
    """Get the actual callable that can be called to obtain the fixture
    value."""
    fixturefunc = fixturedef.func
    # The fixture function needs to be bound to the actual
    # request.instance so that code working with "fixturedef" behaves
    # as expected.
    instance = request.instance
    if instance is not None:
        # Handle the case where fixture is defined not in a test class, but some other class
        # (for example a plugin class with a fixture), see #2270.
        if hasattr(fixturefunc, "__self__") and not isinstance(
            instance,
            fixturefunc.__self__.__class__,
        ):
            return fixturefunc
        fixturefunc = getimfunc(fixturedef.func)
        if fixturefunc != fixturedef.func:
            fixturefunc = fixturefunc.__get__(instance)
    return fixturefunc


def pytest_fixture_setup(
    fixturedef: FixtureDef[FixtureValue], request: SubRequest
) -> FixtureValue:
    """Execution of fixture setup."""
    kwargs = {}
    for argname in fixturedef.argnames:
        kwargs[argname] = request.getfixturevalue(argname)

    fixturefunc = resolve_fixture_function(fixturedef, request)
    my_cache_key = fixturedef.cache_key(request)

    if inspect.isasyncgenfunction(fixturefunc) or inspect.iscoroutinefunction(
        fixturefunc
    ):
        auto_str = " with autouse=True" if fixturedef._autouse else ""

        warnings.warn(
            PytestRemovedIn9Warning(
                f"{request.node.name!r} requested an async fixture "
                f"{request.fixturename!r}{auto_str}, with no plugin or hook that "
                "handled it. This is usually an error, as pytest does not natively "
                "support it. "
                "This will turn into an error in pytest 9.\n"
                "See: https://docs.pytest.org/en/stable/deprecations.html#sync-test-depending-on-async-fixture"
            ),
            # no stacklevel will point at users code, so we just point here
            stacklevel=1,
        )

    try:
        result = call_fixture_func(fixturefunc, request, kwargs)
    except TEST_OUTCOME as e:
        if isinstance(e, skip.Exception):
            # The test requested a fixture which caused a skip.
            # Don't show the fixture as the skip location, as then the user
            # wouldn't know which test skipped.
            e._use_item_location = True
        fixturedef.cached_result = (None, my_cache_key, (e, e.__traceback__))
        raise
    fixturedef.cached_result = (result, my_cache_key, None)
    return result


@final
@dataclasses.dataclass(frozen=True)
class FixtureFunctionMarker:
    scope: _ScopeName | Callable[[str, Config], _ScopeName]
    params: tuple[object, ...] | None
    autouse: bool = False
    ids: tuple[object | None, ...] | Callable[[Any], object | None] | None = None
    name: str | None = None

    _ispytest: dataclasses.InitVar[bool] = False

    def __post_init__(self, _ispytest: bool) -> None:
        check_ispytest(_ispytest)

    def __call__(self, function: FixtureFunction) -> FixtureFunctionDefinition:
        if inspect.isclass(function):
            raise ValueError("class fixtures not supported (maybe in the future)")

        if isinstance(function, FixtureFunctionDefinition):
            raise ValueError(
                f"@pytest.fixture is being applied more than once to the same function {function.__name__!r}"
            )

        if hasattr(function, "pytestmark"):
            warnings.warn(MARKED_FIXTURE, stacklevel=2)

        fixture_definition = FixtureFunctionDefinition(
            function=function, fixture_function_marker=self, _ispytest=True
        )

        name = self.name or function.__name__
        if name == "request":
            location = getlocation(function)
            fail(
                f"'request' is a reserved word for fixtures, use another name:\n  {location}",
                pytrace=False,
            )

        return fixture_definition


# TODO: paramspec/return type annotation tracking and storing
class FixtureFunctionDefinition:
    def __init__(
        self,
        *,
        function: Callable[..., Any],
        fixture_function_marker: FixtureFunctionMarker,
        instance: object | None = None,
        _ispytest: bool = False,
    ) -> None:
        check_ispytest(_ispytest)
        self.name = fixture_function_marker.name or function.__name__
        # In order to show the function that this fixture contains in messages.
        # Set the __name__ to be same as the function __name__ or the given fixture name.
        self.__name__ = self.name
        self._fixture_function_marker = fixture_function_marker
        if instance is not None:
            self._fixture_function = cast(
                Callable[..., Any], function.__get__(instance)
            )
        else:
            self._fixture_function = function
        functools.update_wrapper(self, function)

    def __repr__(self) -> str:
        return f"<pytest_fixture({self._fixture_function})>"

    def __get__(self, instance, owner=None):
        """Behave like a method if the function it was applied to was a method."""
        return FixtureFunctionDefinition(
            function=self._fixture_function,
            fixture_function_marker=self._fixture_function_marker,
            instance=instance,
            _ispytest=True,
        )

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        message = (
            f'Fixture "{self.name}" called directly. Fixtures are not meant to be called directly,\n'
            "but are created automatically when test functions request them as parameters.\n"
            "See https://docs.pytest.org/en/stable/explanation/fixtures.html for more information about fixtures, and\n"
            "https://docs.pytest.org/en/stable/deprecations.html#calling-fixtures-directly"
        )
        fail(message, pytrace=False)

    def _get_wrapped_function(self) -> Callable[..., Any]:
        return self._fixture_function


@overload
def fixture(
    fixture_function: Callable[..., object],
    *,
    scope: _ScopeName | Callable[[str, Config], _ScopeName] = ...,
    params: Iterable[object] | None = ...,
    autouse: bool = ...,
    ids: Sequence[object | None] | Callable[[Any], object | None] | None = ...,
    name: str | None = ...,
) -> FixtureFunctionDefinition: ...


@overload
def fixture(
    fixture_function: None = ...,
    *,
    scope: _ScopeName | Callable[[str, Config], _ScopeName] = ...,
    params: Iterable[object] | None = ...,
    autouse: bool = ...,
    ids: Sequence[object | None] | Callable[[Any], object | None] | None = ...,
    name: str | None = None,
) -> FixtureFunctionMarker: ...


def fixture(
    fixture_function: FixtureFunction | None = None,
    *,
    scope: _ScopeName | Callable[[str, Config], _ScopeName] = "function",
    params: Iterable[object] | None = None,
    autouse: bool = False,
    ids: Sequence[object | None] | Callable[[Any], object | None] | None = None,
    name: str | None = None,
) -> FixtureFunctionMarker | FixtureFunctionDefinition:
    """Decorator to mark a fixture factory function.

    This decorator can be used, with or without parameters, to define a
    fixture function.

    The name of the fixture function can later be referenced to cause its
    invocation ahead of running tests: test modules or classes can use the
    ``pytest.mark.usefixtures(fixturename)`` marker.

    Test functions can directly use fixture names as input arguments in which
    case the fixture instance returned from the fixture function will be
    injected.

    Fixtures can provide their values to test functions using ``return`` or
    ``yield`` statements. When using ``yield`` the code block after the
    ``yield`` statement is executed as teardown code regardless of the test
    outcome, and must yield exactly once.

    :param scope:
        The scope for which this fixture is shared; one of ``"function"``
        (default), ``"class"``, ``"module"``, ``"package"`` or ``"session"``.

        This parameter may also be a callable which receives ``(fixture_name, config)``
        as parameters, and must return a ``str`` with one of the values mentioned above.

        See :ref:`dynamic scope` in the docs for more information.

    :param params:
        An optional list of parameters which will cause multiple invocations
        of the fixture function and all of the tests using it. The current
        parameter is available in ``request.param``.

    :param autouse:
        If True, the fixture func is activated for all tests that can see it.
        If False (the default), an explicit reference is needed to activate
        the fixture.

    :param ids:
        Sequence of ids each corresponding to the params so that they are
        part of the test id. If no ids are provided they will be generated
        automatically from the params.

    :param name:
        The name of the fixture. This defaults to the name of the decorated
        function. If a fixture is used in the same module in which it is
        defined, the function name of the fixture will be shadowed by the
        function arg that requests the fixture; one way to resolve this is to
        name the decorated function ``fixture_<fixturename>`` and then use
        ``@pytest.fixture(name='<fixturename>')``.
    """
    fixture_marker = FixtureFunctionMarker(
        scope=scope,
        params=tuple(params) if params is not None else None,
        autouse=autouse,
        ids=None if ids is None else ids if callable(ids) else tuple(ids),
        name=name,
        _ispytest=True,
    )

    # Direct decoration.
    if fixture_function:
        return fixture_marker(fixture_function)

    return fixture_marker


def yield_fixture(
    fixture_function=None,
    *args,
    scope="function",
    params=None,
    autouse=False,
    ids=None,
    name=None,
):
    """(Return a) decorator to mark a yield-fixture factory function.

    .. deprecated:: 3.0
        Use :py:func:`pytest.fixture` directly instead.
    """
    warnings.warn(YIELD_FIXTURE, stacklevel=2)
    return fixture(
        fixture_function,
        *args,
        scope=scope,
        params=params,
        autouse=autouse,
        ids=ids,
        name=name,
    )


@fixture(scope="session")
def pytestconfig(request: FixtureRequest) -> Config:
    """Session-scoped fixture that returns the session's :class:`pytest.Config`
    object.

    Example::

        def test_foo(pytestconfig):
            if pytestconfig.get_verbosity() > 0:
                ...

    """
    return request.config


def pytest_addoption(parser: Parser) -> None:
    parser.addini(
        "usefixtures",
        type="args",
        default=[],
        help="List of default fixtures to be used with this project",
    )
    group = parser.getgroup("general")
    group.addoption(
        "--fixtures",
        "--funcargs",
        action="store_true",
        dest="showfixtures",
        default=False,
        help="Show available fixtures, sorted by plugin appearance "
        "(fixtures with leading '_' are only shown with '-v')",
    )
    group.addoption(
        "--fixtures-per-test",
        action="store_true",
        dest="show_fixtures_per_test",
        default=False,
        help="Show fixtures per test",
    )


def pytest_cmdline_main(config: Config) -> int | ExitCode | None:
    if config.option.showfixtures:
        showfixtures(config)
        return 0
    if config.option.show_fixtures_per_test:
        show_fixtures_per_test(config)
        return 0
    return None


def _get_direct_parametrize_args(node: nodes.Node) -> set[str]:
    """Return all direct parametrization arguments of a node, so we don't
    mistake them for fixtures.

    Check https://github.com/pytest-dev/pytest/issues/5036.

    These things are done later as well when dealing with parametrization
    so this could be improved.
    """
    parametrize_argnames: set[str] = set()
    for marker in node.iter_markers(name="parametrize"):
        if not marker.kwargs.get("indirect", False):
            p_argnames, _ = ParameterSet._parse_parametrize_args(
                *marker.args, **marker.kwargs
            )
            parametrize_argnames.update(p_argnames)
    return parametrize_argnames


def deduplicate_names(*seqs: Iterable[str]) -> tuple[str, ...]:
    """De-duplicate the sequence of names while keeping the original order."""
    # Ideally we would use a set, but it does not preserve insertion order.
    return tuple(dict.fromkeys(name for seq in seqs for name in seq))


class FixtureManager:
    """pytest fixture definitions and information is stored and managed
    from this class.

    During collection fm.parsefactories() is called multiple times to parse
    fixture function definitions into FixtureDef objects and internal
    data structures.

    During collection of test functions, metafunc-mechanics instantiate
    a FuncFixtureInfo object which is cached per node/func-name.
    This FuncFixtureInfo object is later retrieved by Function nodes
    which themselves offer a fixturenames attribute.

    The FuncFixtureInfo object holds information about fixtures and FixtureDefs
    relevant for a particular function. An initial list of fixtures is
    assembled like this:

    - ini-defined usefixtures
    - autouse-marked fixtures along the collection chain up from the function
    - usefixtures markers at module/class/function level
    - test function funcargs

    Subsequently the funcfixtureinfo.fixturenames attribute is computed
    as the closure of the fixtures needed to setup the initial fixtures,
    i.e. fixtures needed by fixture functions themselves are appended
    to the fixturenames list.

    Upon the test-setup phases all fixturenames are instantiated, retrieved
    by a lookup of their FuncFixtureInfo.
    """

    def __init__(self, session: Session) -> None:
        self.session = session
        self.config: Config = session.config
        # Maps a fixture name (argname) to all of the FixtureDefs in the test
        # suite/plugins defined with this name. Populated by parsefactories().
        # TODO: The order of the FixtureDefs list of each arg is significant,
        #       explain.
        self._arg2fixturedefs: Final[dict[str, list[FixtureDef[Any]]]] = {}
        self._holderobjseen: Final[set[object]] = set()
        # A mapping from a nodeid to a list of autouse fixtures it defines.
        self._nodeid_autousenames: Final[dict[str, list[str]]] = {
            "": self.config.getini("usefixtures"),
        }
        session.config.pluginmanager.register(self, "funcmanage")

    def getfixtureinfo(
        self,
        node: nodes.Item,
        func: Callable[..., object] | None,
        cls: type | None,
    ) -> FuncFixtureInfo:
        """Calculate the :class:`FuncFixtureInfo` for an item.

        If ``func`` is None, or if the item sets an attribute
        ``nofuncargs = True``, then ``func`` is not examined at all.

        :param node:
            The item requesting the fixtures.
        :param func:
            The item's function.
        :param cls:
            If the function is a method, the method's class.
        """
        if func is not None and not getattr(node, "nofuncargs", False):
            argnames = getfuncargnames(func, name=node.name, cls=cls)
        else:
            argnames = ()
        usefixturesnames = self._getusefixturesnames(node)
        autousenames = self._getautousenames(node)
        initialnames = deduplicate_names(autousenames, usefixturesnames, argnames)

        direct_parametrize_args = _get_direct_parametrize_args(node)

        names_closure, arg2fixturedefs = self.getfixtureclosure(
            parentnode=node,
            initialnames=initialnames,
            ignore_args=direct_parametrize_args,
        )

        return FuncFixtureInfo(argnames, initialnames, names_closure, arg2fixturedefs)

    def pytest_plugin_registered(self, plugin: _PluggyPlugin, plugin_name: str) -> None:
        # Fixtures defined in conftest plugins are only visible to within the
        # conftest's directory. This is unlike fixtures in non-conftest plugins
        # which have global visibility. So for conftests, construct the base
        # nodeid from the plugin name (which is the conftest path).
        if plugin_name and plugin_name.endswith("conftest.py"):
            # Note: we explicitly do *not* use `plugin.__file__` here -- The
            # difference is that plugin_name has the correct capitalization on
            # case-insensitive systems (Windows) and other normalization issues
            # (issue #11816).
            conftestpath = absolutepath(plugin_name)
            try:
                nodeid = str(conftestpath.parent.relative_to(self.config.rootpath))
            except ValueError:
                nodeid = ""
            if nodeid == ".":
                nodeid = ""
            if os.sep != nodes.SEP:
                nodeid = nodeid.replace(os.sep, nodes.SEP)
        else:
            nodeid = None

        self.parsefactories(plugin, nodeid)

    def _getautousenames(self, node: nodes.Node) -> Iterator[str]:
        """Return the names of autouse fixtures applicable to node."""
        for parentnode in node.listchain():
            basenames = self._nodeid_autousenames.get(parentnode.nodeid)
            if basenames:
                yield from basenames

    def _getusefixturesnames(self, node: nodes.Item) -> Iterator[str]:
        """Return the names of usefixtures fixtures applicable to node."""
        for marker_node, mark in node.iter_markers_with_node(name="usefixtures"):
            if not mark.args:
                marker_node.warn(
                    PytestWarning(
                        f"usefixtures() in {node.nodeid} without arguments has no effect"
                    )
                )
            yield from mark.args

    def getfixtureclosure(
        self,
        parentnode: nodes.Node,
        initialnames: tuple[str, ...],
        ignore_args: AbstractSet[str],
    ) -> tuple[list[str], dict[str, Sequence[FixtureDef[Any]]]]:
        # Collect the closure of all fixtures, starting with the given
        # fixturenames as the initial set.  As we have to visit all
        # factory definitions anyway, we also return an arg2fixturedefs
        # mapping so that the caller can reuse it and does not have
        # to re-discover fixturedefs again for each fixturename
        # (discovering matching fixtures for a given name/node is expensive).

        fixturenames_closure = list(initialnames)

        arg2fixturedefs: dict[str, Sequence[FixtureDef[Any]]] = {}
        lastlen = -1
        while lastlen != len(fixturenames_closure):
            lastlen = len(fixturenames_closure)
            for argname in fixturenames_closure:
                if argname in ignore_args:
                    continue
                if argname in arg2fixturedefs:
                    continue
                fixturedefs = self.getfixturedefs(argname, parentnode)
                if fixturedefs:
                    arg2fixturedefs[argname] = fixturedefs
                    for arg in fixturedefs[-1].argnames:
                        if arg not in fixturenames_closure:
                            fixturenames_closure.append(arg)

        def sort_by_scope(arg_name: str) -> Scope:
            try:
                fixturedefs = arg2fixturedefs[arg_name]
            except KeyError:
                return Scope.Function
            else:
                return fixturedefs[-1]._scope

        fixturenames_closure.sort(key=sort_by_scope, reverse=True)
        return fixturenames_closure, arg2fixturedefs

    def pytest_generate_tests(self, metafunc: Metafunc) -> None:
        """Generate new tests based on parametrized fixtures used by the given metafunc"""

        def get_parametrize_mark_argnames(mark: Mark) -> Sequence[str]:
            args, _ = ParameterSet._parse_parametrize_args(*mark.args, **mark.kwargs)
            return args

        for argname in metafunc.fixturenames:
            # Get the FixtureDefs for the argname.
            fixture_defs = metafunc._arg2fixturedefs.get(argname)
            if not fixture_defs:
                # Will raise FixtureLookupError at setup time if not parametrized somewhere
                # else (e.g @pytest.mark.parametrize)
                continue

            # If the test itself parametrizes using this argname, give it
            # precedence.
            if any(
                argname in get_parametrize_mark_argnames(mark)
                for mark in metafunc.definition.iter_markers("parametrize")
            ):
                continue

            # In the common case we only look at the fixture def with the
            # closest scope (last in the list). But if the fixture overrides
            # another fixture, while requesting the super fixture, keep going
            # in case the super fixture is parametrized (#1953).
            for fixturedef in reversed(fixture_defs):
                # Fixture is parametrized, apply it and stop.
                if fixturedef.params is not None:
                    metafunc.parametrize(
                        argname,
                        fixturedef.params,
                        indirect=True,
                        scope=fixturedef.scope,
                        ids=fixturedef.ids,
                    )
                    break

                # Not requesting the overridden super fixture, stop.
                if argname not in fixturedef.argnames:
                    break

                # Try next super fixture, if any.

    def pytest_collection_modifyitems(self, items: list[nodes.Item]) -> None:
        # Separate parametrized setups.
        items[:] = reorder_items(items)

    def _register_fixture(
        self,
        *,
        name: str,
        func: _FixtureFunc[object],
        nodeid: str | None,
        scope: Scope | _ScopeName | Callable[[str, Config], _ScopeName] = "function",
        params: Sequence[object] | None = None,
        ids: tuple[object | None, ...] | Callable[[Any], object | None] | None = None,
        autouse: bool = False,
    ) -> None:
        """Register a fixture

        :param name:
            The fixture's name.
        :param func:
            The fixture's implementation function.
        :param nodeid:
            The visibility of the fixture. The fixture will be available to the
            node with this nodeid and its children in the collection tree.
            None means that the fixture is visible to the entire collection tree,
            e.g. a fixture defined for general use in a plugin.
        :param scope:
            The fixture's scope.
        :param params:
            The fixture's parametrization params.
        :param ids:
            The fixture's IDs.
        :param autouse:
            Whether this is an autouse fixture.
        """
        fixture_def = FixtureDef(
            config=self.config,
            baseid=nodeid,
            argname=name,
            func=func,
            scope=scope,
            params=params,
            ids=ids,
            _ispytest=True,
            _autouse=autouse,
        )

        faclist = self._arg2fixturedefs.setdefault(name, [])
        if fixture_def.has_location:
            faclist.append(fixture_def)
        else:
            # fixturedefs with no location are at the front
            # so this inserts the current fixturedef after the
            # existing fixturedefs from external plugins but
            # before the fixturedefs provided in conftests.
            i = len([f for f in faclist if not f.has_location])
            faclist.insert(i, fixture_def)
        if autouse:
            self._nodeid_autousenames.setdefault(nodeid or "", []).append(name)

    @overload
    def parsefactories(
        self,
        node_or_obj: nodes.Node,
    ) -> None:
        raise NotImplementedError()

    @overload
    def parsefactories(
        self,
        node_or_obj: object,
        nodeid: str | None,
    ) -> None:
        raise NotImplementedError()

    def parsefactories(
        self,
        node_or_obj: nodes.Node | object,
        nodeid: str | NotSetType | None = NOTSET,
    ) -> None:
        """Collect fixtures from a collection node or object.

        Found fixtures are parsed into `FixtureDef`s and saved.

        If `node_or_object` is a collection node (with an underlying Python
        object), the node's object is traversed and the node's nodeid is used to
        determine the fixtures' visibility. `nodeid` must not be specified in
        this case.

        If `node_or_object` is an object (e.g. a plugin), the object is
        traversed and the given `nodeid` is used to determine the fixtures'
        visibility. `nodeid` must be specified in this case; None and "" mean
        total visibility.
        """
        if nodeid is not NOTSET:
            holderobj = node_or_obj
        else:
            assert isinstance(node_or_obj, nodes.Node)
            holderobj = cast(object, node_or_obj.obj)  # type: ignore[attr-defined]
            assert isinstance(node_or_obj.nodeid, str)
            nodeid = node_or_obj.nodeid
        if holderobj in self._holderobjseen:
            return

        # Avoid accessing `@property` (and other descriptors) when iterating fixtures.
        if not safe_isclass(holderobj) and not isinstance(holderobj, types.ModuleType):
            holderobj_tp: object = type(holderobj)
        else:
            holderobj_tp = holderobj

        self._holderobjseen.add(holderobj)
        for name in dir(holderobj):
            # The attribute can be an arbitrary descriptor, so the attribute
            # access below can raise. safe_getattr() ignores such exceptions.
            obj_ub = safe_getattr(holderobj_tp, name, None)
            if type(obj_ub) is FixtureFunctionDefinition:
                marker = obj_ub._fixture_function_marker
                if marker.name:
                    fixture_name = marker.name
                else:
                    fixture_name = name

                # OK we know it is a fixture -- now safe to look up on the _instance_.
                try:
                    obj = getattr(holderobj, name)
                # if the fixture is named in the decorator we cannot find it in the module
                except AttributeError:
                    obj = obj_ub

                func = obj._get_wrapped_function()

                self._register_fixture(
                    name=fixture_name,
                    nodeid=nodeid,
                    func=func,
                    scope=marker.scope,
                    params=marker.params,
                    ids=marker.ids,
                    autouse=marker.autouse,
                )

    def getfixturedefs(
        self, argname: str, node: nodes.Node
    ) -> Sequence[FixtureDef[Any]] | None:
        """Get FixtureDefs for a fixture name which are applicable
        to a given node.

        Returns None if there are no fixtures at all defined with the given
        name. (This is different from the case in which there are fixtures
        with the given name, but none applicable to the node. In this case,
        an empty result is returned).

        :param argname: Name of the fixture to search for.
        :param node: The requesting Node.
        """
        try:
            fixturedefs = self._arg2fixturedefs[argname]
        except KeyError:
            return None
        return tuple(self._matchfactories(fixturedefs, node))

    def _matchfactories(
        self, fixturedefs: Iterable[FixtureDef[Any]], node: nodes.Node
    ) -> Iterator[FixtureDef[Any]]:
        parentnodeids = {n.nodeid for n in node.iter_parents()}
        for fixturedef in fixturedefs:
            if fixturedef.baseid in parentnodeids:
                yield fixturedef


def show_fixtures_per_test(config: Config) -> int | ExitCode:
    from _pytest.main import wrap_session

    return wrap_session(config, _show_fixtures_per_test)


_PYTEST_DIR = Path(_pytest.__file__).parent


def _pretty_fixture_path(invocation_dir: Path, func) -> str:
    loc = Path(getlocation(func, invocation_dir))
    prefix = Path("...", "_pytest")
    try:
        return str(prefix / loc.relative_to(_PYTEST_DIR))
    except ValueError:
        return bestrelpath(invocation_dir, loc)


def _show_fixtures_per_test(config: Config, session: Session) -> None:
    import _pytest.config

    session.perform_collect()
    invocation_dir = config.invocation_params.dir
    tw = _pytest.config.create_terminal_writer(config)
    verbose = config.get_verbosity()

    def get_best_relpath(func) -> str:
        loc = getlocation(func, invocation_dir)
        return bestrelpath(invocation_dir, Path(loc))

    def write_fixture(fixture_def: FixtureDef[object]) -> None:
        argname = fixture_def.argname
        if verbose <= 0 and argname.startswith("_"):
            return
        prettypath = _pretty_fixture_path(invocation_dir, fixture_def.func)
        tw.write(f"{argname}", green=True)
        tw.write(f" -- {prettypath}", yellow=True)
        tw.write("\n")
        fixture_doc = inspect.getdoc(fixture_def.func)
        if fixture_doc:
            write_docstring(
                tw,
                fixture_doc.split("\n\n", maxsplit=1)[0]
                if verbose <= 0
                else fixture_doc,
            )
        else:
            tw.line("    no docstring available", red=True)

    def write_item(item: nodes.Item) -> None:
        # Not all items have _fixtureinfo attribute.
        info: FuncFixtureInfo | None = getattr(item, "_fixtureinfo", None)
        if info is None or not info.name2fixturedefs:
            # This test item does not use any fixtures.
            return
        tw.line()
        tw.sep("-", f"fixtures used by {item.name}")
        # TODO: Fix this type ignore.
        tw.sep("-", f"({get_best_relpath(item.function)})")  # type: ignore[attr-defined]
        # dict key not used in loop but needed for sorting.
        for _, fixturedefs in sorted(info.name2fixturedefs.items()):
            assert fixturedefs is not None
            if not fixturedefs:
                continue
            # Last item is expected to be the one used by the test item.
            write_fixture(fixturedefs[-1])

    for session_item in session.items:
        write_item(session_item)


def showfixtures(config: Config) -> int | ExitCode:
    from _pytest.main import wrap_session

    return wrap_session(config, _showfixtures_main)


def _showfixtures_main(config: Config, session: Session) -> None:
    import _pytest.config

    session.perform_collect()
    invocation_dir = config.invocation_params.dir
    tw = _pytest.config.create_terminal_writer(config)
    verbose = config.get_verbosity()

    fm = session._fixturemanager

    available = []
    seen: set[tuple[str, str]] = set()

    for argname, fixturedefs in fm._arg2fixturedefs.items():
        assert fixturedefs is not None
        if not fixturedefs:
            continue
        for fixturedef in fixturedefs:
            loc = getlocation(fixturedef.func, invocation_dir)
            if (fixturedef.argname, loc) in seen:
                continue
            seen.add((fixturedef.argname, loc))
            available.append(
                (
                    len(fixturedef.baseid),
                    fixturedef.func.__module__,
                    _pretty_fixture_path(invocation_dir, fixturedef.func),
                    fixturedef.argname,
                    fixturedef,
                )
            )

    available.sort()
    currentmodule = None
    for baseid, module, prettypath, argname, fixturedef in available:
        if currentmodule != module:
            if not module.startswith("_pytest."):
                tw.line()
                tw.sep("-", f"fixtures defined from {module}")
                currentmodule = module
        if verbose <= 0 and argname.startswith("_"):
            continue
        tw.write(f"{argname}", green=True)
        if fixturedef.scope != "function":
            tw.write(f" [{fixturedef.scope} scope]", cyan=True)
        tw.write(f" -- {prettypath}", yellow=True)
        tw.write("\n")
        doc = inspect.getdoc(fixturedef.func)
        if doc:
            write_docstring(
                tw, doc.split("\n\n", maxsplit=1)[0] if verbose <= 0 else doc
            )
        else:
            tw.line("    no docstring available", red=True)
        tw.line()


def write_docstring(tw: TerminalWriter, doc: str, indent: str = "    ") -> None:
    for line in doc.split("\n"):
        tw.line(indent + line)



"""
Cloud Browser Yer İmi Yöneticisi
JSON tabanlı yer imi sistemi
"""

import json
import os
from datetime import datetime
from typing import List, Dict, Any, Optional
from pathlib import Path

try:
    from PyQt6.QtWidgets import *
    from PyQt6.QtCore import *
    from PyQt6.QtGui import *
except ImportError:
    raise ImportError("PyQt6 is required for Cloud Browser")

class BookmarkManager:
    """
    Yer imi yöneticisi sınıfı
    """
    
    def __init__(self):
        self.bookmarks: List[Dict[str, Any]] = []
        self.bookmarks_file = Path.home() / ".cloud_browser" / "bookmarks.json"
        self.bookmarks_file.parent.mkdir(exist_ok=True)
        self.load_bookmarks()
    
    def load_bookmarks(self):
        """Yer imlerini dosyadan yükle"""
        try:
            if self.bookmarks_file.exists():
                with open(self.bookmarks_file, 'r', encoding='utf-8') as f:
                    data = json.load(f)
                    self.bookmarks = data.get('bookmarks', [])
            else:
                # Varsayılan yer imleri
                self.bookmarks = [
                    {
                        "title": "Google",
                        "url": "https://www.google.com",
                        "folder": "Genel",
                        "created": datetime.now().isoformat(),
                        "visits": 0,
                        "favicon": "🔍"
                    },
                    {
                        "title": "YouTube",
                        "url": "https://www.youtube.com",
                        "folder": "Genel",
                        "created": datetime.now().isoformat(),
                        "visits": 0,
                        "favicon": "📺"
                    },
                    {
                        "title": "GitHub",
                        "url": "https://www.github.com",
                        "folder": "Geliştirme",
                        "created": datetime.now().isoformat(),
                        "visits": 0,
                        "favicon": "💻"
                    }
                ]
                self.save_bookmarks()
        except Exception as e:
            print(f"Yer imi yükleme hatası: {e}")
            self.bookmarks = []
    
    def save_bookmarks(self):
        """Yer imlerini dosyaya kaydet"""
        try:
            data = {
                "version": "2.0.0",
                "created": datetime.now().isoformat(),
                "bookmarks": self.bookmarks
            }
            with open(self.bookmarks_file, 'w', encoding='utf-8') as f:
                json.dump(data, f, indent=2, ensure_ascii=False, default=str)
        except Exception as e:
            print(f"Yer imi kaydetme hatası: {e}")
    
    def add_bookmark(self, title: str, url: str, folder: str = "Genel", favicon: str = "🌐"):
        """Yer imi ekle"""
        # Zaten var mı kontrol et
        if self.is_bookmarked(url):
            return False
        
        bookmark = {
            "title": title.strip(),
            "url": url.strip(),
            "folder": folder.strip(),
            "created": datetime.now().isoformat(),
            "visits": 0,
            "favicon": favicon
        }
        
        self.bookmarks.append(bookmark)
        self.save_bookmarks()
        return True
    
    def remove_bookmark(self, url: str):
        """Yer imi sil"""
        original_count = len(self.bookmarks)
        self.bookmarks = [b for b in self.bookmarks if b["url"] != url]
        
        if len(self.bookmarks) < original_count:
            self.save_bookmarks()
            return True
        return False
    
    def update_bookmark(self, old_url: str, title: str, url: str, folder: str):
        """Yer imi güncelle"""
        for bookmark in self.bookmarks:
            if bookmark["url"] == old_url:
                bookmark["title"] = title.strip()
                bookmark["url"] = url.strip()
                bookmark["folder"] = folder.strip()
                bookmark["modified"] = datetime.now().isoformat()
                self.save_bookmarks()
                return True
        return False
    
    def get_bookmarks(self, folder: Optional[str] = None) -> List[Dict[str, Any]]:
        """Yer imlerini al"""
        if folder:
            return [b for b in self.bookmarks if b.get("folder") == folder]
        return self.bookmarks.copy()
    
    def get_folders(self) -> List[str]:
        """Klasör listesini al"""
        folders = set()
        for bookmark in self.bookmarks:
            folders.add(bookmark.get("folder", "Genel"))
        return sorted(list(folders))
    
    def is_bookmarked(self, url: str) -> bool:
        """URL yer imi var mı?"""
        return any(b["url"] == url for b in self.bookmarks)
    
    def increment_visits(self, url: str):
        """Ziyaret sayısını artır"""
        for bookmark in self.bookmarks:
            if bookmark["url"] == url:
                bookmark["visits"] = bookmark.get("visits", 0) + 1
                bookmark["last_visited"] = datetime.now().isoformat()
                self.save_bookmarks()
                break
    
    def search_bookmarks(self, query: str) -> List[Dict[str, Any]]:
        """Yer imlerinde arama yap"""
        query = query.lower().strip()
        if not query:
            return self.bookmarks.copy()
        
        results = []
        for bookmark in self.bookmarks:
            title = bookmark.get("title", "").lower()
            url = bookmark.get("url", "").lower()
            folder = bookmark.get("folder", "").lower()
            
            if query in title or query in url or query in folder:
                results.append(bookmark)
        
        return results

class BookmarkManagerDialog(QDialog):
    """
    Yer imi yöneticisi dialog'u
    """
    
    def __init__(self, bookmark_manager: BookmarkManager, parent=None):
        super().__init__(parent)
        self.bookmark_manager = bookmark_manager
        self.parent_browser = parent
        
        self.setWindowTitle("Yer İmi Yöneticisi")
        self.setModal(True)
        self.resize(800, 600)
        
        self.init_ui()
        self.load_bookmarks()
    
    def init_ui(self):
        """UI'yı başlat"""
        layout = QVBoxLayout(self)
        
        # Üst toolbar
        toolbar_layout = QHBoxLayout()
        
        # Arama kutusu
        self.search_box = QLineEdit()
        self.search_box.setPlaceholderText("Yer imlerinde ara...")
        self.search_box.textChanged.connect(self.filter_bookmarks)
        toolbar_layout.addWidget(QLabel("🔍"))
        toolbar_layout.addWidget(self.search_box)
        
        # Klasör filtresi
        self.folder_combo = QComboBox()
        self.folder_combo.addItem("Tüm Klasörler")
        self.folder_combo.currentTextChanged.connect(self.filter_bookmarks)
        toolbar_layout.addWidget(QLabel("📁"))
        toolbar_layout.addWidget(self.folder_combo)
        
        # Spacer
        toolbar_layout.addStretch()
        
        # Yeni yer imi butonu
        add_btn = QPushButton("➕ Yeni Yer İmi")
        add_btn.clicked.connect(self.add_bookmark)
        toolbar_layout.addWidget(add_btn)
        
        layout.addLayout(toolbar_layout)
        
        # Yer imi listesi
        self.bookmark_list = QTreeWidget()
        self.bookmark_list.setHeaderLabels(["Başlık", "URL", "Klasör", "Oluşturulma", "Ziyaret"])
        self.bookmark_list.setRootIsDecorated(False)
        self.bookmark_list.setAlternatingRowColors(True)
        self.bookmark_list.setSortingEnabled(True)
        self.bookmark_list.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        
        # Context menu
        self.bookmark_list.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.bookmark_list.customContextMenuRequested.connect(self.show_context_menu)
        
        # Double click
        self.bookmark_list.itemDoubleClicked.connect(self.open_bookmark)
        
        layout.addWidget(self.bookmark_list)
        
        # Alt butonlar
        button_layout = QHBoxLayout()
        
        open_btn = QPushButton("🌐 Aç")
        open_btn.clicked.connect(self.open_bookmark)
        button_layout.addWidget(open_btn)
        
        edit_btn = QPushButton("✏️ Düzenle")
        edit_btn.clicked.connect(self.edit_bookmark)
        button_layout.addWidget(edit_btn)
        
        delete_btn = QPushButton("🗑️ Sil")
        delete_btn.clicked.connect(self.delete_bookmark)
        button_layout.addWidget(delete_btn)
        
        button_layout.addStretch()
        
        export_btn = QPushButton("📤 Dışa Aktar")
        export_btn.clicked.connect(self.export_bookmarks)
        button_layout.addWidget(export_btn)
        
        import_btn = QPushButton("📥 İçe Aktar")
        import_btn.clicked.connect(self.import_bookmarks)
        button_layout.addWidget(import_btn)
        
        button_layout.addStretch()
        
        close_btn = QPushButton("❌ Kapat")
        close_btn.clicked.connect(self.close)
        button_layout.addWidget(close_btn)
        
        layout.addLayout(button_layout)
        
        # Stil uygula
        self.setStyleSheet("""
            QDialog {
                background-color: #f8f9fa;
            }
            QTreeWidget {
                background-color: white;
                border: 1px solid #dee2e6;
                border-radius: 6px;
                font-size: 13px;
            }
            QTreeWidget::item {
                padding: 8px;
                border-bottom: 1px solid #f0f0f0;
            }
            QTreeWidget::item:selected {
                background-color: #007bff;
                color: white;
            }
            QTreeWidget::item:hover {
                background-color: #e9ecef;
            }
            QPushButton {
                background-color: #007bff;
                color: white;
                border: none;
                border-radius: 6px;
                padding: 8px 16px;
                font-weight: bold;
            }
            QPushButton:hover {
                background-color: #0056b3;
            }
            QPushButton:pressed {
                background-color: #004085;
            }
            QLineEdit, QComboBox {
                border: 2px solid #dee2e6;
                border-radius: 6px;
                padding: 8px;
                font-size: 13px;
            }
            QLineEdit:focus, QComboBox:focus {
                border-color: #007bff;
            }
        """)
    
    def load_bookmarks(self):
        """Yer imlerini yükle"""
        self.bookmark_list.clear()
        
        # Klasör combo'yu güncelle
        self.folder_combo.clear()
        self.folder_combo.addItem("Tüm Klasörler")
        folders = self.bookmark_manager.get_folders()
        self.folder_combo.addItems(folders)
        
        # Yer imlerini ekle
        bookmarks = self.bookmark_manager.get_bookmarks()
        for bookmark in bookmarks:
            self.add_bookmark_item(bookmark)
        
        # Sütun genişliklerini ayarla
        self.bookmark_list.resizeColumnToContents(0)
        self.bookmark_list.resizeColumnToContents(2)
        self.bookmark_list.resizeColumnToContents(3)
        self.bookmark_list.resizeColumnToContents(4)
    
    def add_bookmark_item(self, bookmark: Dict[str, Any]):
        """Yer imi öğesi ekle"""
        item = QTreeWidgetItem()
        
        # Favicon + başlık
        favicon = bookmark.get("favicon", "🌐")
        title = bookmark.get("title", "")
        item.setText(0, f"{favicon} {title}")
        
        # URL
        item.setText(1, bookmark.get("url", ""))
        
        # Klasör
        item.setText(2, bookmark.get("folder", "Genel"))
        
        # Oluşturulma tarihi
        created = bookmark.get("created", "")
        if created:
            try:
                dt = datetime.fromisoformat(created.replace('Z', '+00:00'))
                item.setText(3, dt.strftime("%d.%m.%Y"))
            except:
                item.setText(3, created[:10])
        
        # Ziyaret sayısı
        visits = bookmark.get("visits", 0)
        item.setText(4, str(visits))
        
        # Data olarak bookmark'ı sakla
        item.setData(0, Qt.ItemDataRole.UserRole, bookmark)
        
        self.bookmark_list.addTopLevelItem(item)
    
    def filter_bookmarks(self):
        """Yer imlerini filtrele"""
        search_text = self.search_box.text().lower()
        selected_folder = self.folder_combo.currentText()
        
        for i in range(self.bookmark_list.topLevelItemCount()):
            item = self.bookmark_list.topLevelItem(i)
            bookmark = item.data(0, Qt.ItemDataRole.UserRole)
            
            # Arama filtresi
            title_match = search_text in bookmark.get("title", "").lower()
            url_match = search_text in bookmark.get("url", "").lower()
            search_match = not search_text or title_match or url_match
            
            # Klasör filtresi
            folder_match = (selected_folder == "Tüm Klasörler" or 
                          bookmark.get("folder", "Genel") == selected_folder)
            
            # Görünürlük
            item.setHidden(not (search_match and folder_match))
    
    def show_context_menu(self, position):
        """Context menü göster"""
        item = self.bookmark_list.itemAt(position)
        if not item:
            return
        
        menu = QMenu(self)
        
        open_action = menu.addAction("🌐 Aç")
        open_action.triggered.connect(self.open_bookmark)
        
        open_new_tab_action = menu.addAction("🆕 Yeni Sekmede Aç")
        open_new_tab_action.triggered.connect(self.open_bookmark_new_tab)
        
        menu.addSeparator()
        
        edit_action = menu.addAction("✏️ Düzenle")
        edit_action.triggered.connect(self.edit_bookmark)
        
        delete_action = menu.addAction("🗑️ Sil")
        delete_action.triggered.connect(self.delete_bookmark)
        
        menu.exec(self.bookmark_list.mapToGlobal(position))
    
    def open_bookmark(self):
        """Yer imini aç"""
        current_item = self.bookmark_list.currentItem()
        if not current_item:
            return
        
        bookmark = current_item.data(0, Qt.ItemDataRole.UserRole)
        url = bookmark.get("url", "")
        
        if url and self.parent_browser:
            # Mevcut sekmede aç
            if hasattr(self.parent_browser, 'tab_widget'):
                current_web_view = self.parent_browser.tab_widget.current_web_view()
                if current_web_view and hasattr(current_web_view, 'setUrl'):
                    current_web_view.setUrl(QUrl(url))
            
            # Ziyaret sayısını artır
            self.bookmark_manager.increment_visits(url)
            
            # Dialog'u kapat
            self.close()
    
    def open_bookmark_new_tab(self):
        """Yer imini yeni sekmede aç"""
        current_item = self.bookmark_list.currentItem()
        if not current_item:
            return
        
        bookmark = current_item.data(0, Qt.ItemDataRole.UserRole)
        url = bookmark.get("url", "")
        
        if url and self.parent_browser:
            # Yeni sekmede aç
            self.parent_browser.new_tab(url)
            
            # Ziyaret sayısını artır
            self.bookmark_manager.increment_visits(url)
    
    def add_bookmark(self):
        """Yeni yer imi ekle"""
        dialog = BookmarkEditDialog(self.bookmark_manager, parent=self)
        if dialog.exec() == QDialog.DialogCode.Accepted:
            self.load_bookmarks()
    
    def edit_bookmark(self):
        """Yer imi düzenle"""
        current_item = self.bookmark_list.currentItem()
        if not current_item:
            return
        
        bookmark = current_item.data(0, Qt.ItemDataRole.UserRole)
        dialog = BookmarkEditDialog(self.bookmark_manager, bookmark, self)
        if dialog.exec() == QDialog.DialogCode.Accepted:
            self.load_bookmarks()
    
    def delete_bookmark(self):
        """Yer imi sil"""
        selected_items = self.bookmark_list.selectedItems()
        if not selected_items:
            return
        
        reply = QMessageBox.question(
            self,
            "Yer İmi Sil",
            f"{len(selected_items)} yer imi silinecek. Emin misiniz?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
        )
        
        if reply == QMessageBox.StandardButton.Yes:
            for item in selected_items:
                bookmark = item.data(0, Qt.ItemDataRole.UserRole)
                url = bookmark.get("url", "")
                self.bookmark_manager.remove_bookmark(url)
            
            self.load_bookmarks()
    
    def export_bookmarks(self):
        """Yer imlerini dışa aktar"""
        file_path, _ = QFileDialog.getSaveFileName(
            self,
            "Yer İmlerini Dışa Aktar",
            "bookmarks.json",
            "JSON Dosyaları (*.json);;Tüm Dosyalar (*.*)"
        )
        
        if file_path:
            try:
                data = {
                    "version": "2.0.0",
                    "exported": datetime.now().isoformat(),
                    "bookmarks": self.bookmark_manager.get_bookmarks()
                }
                with open(file_path, 'w', encoding='utf-8') as f:
                    json.dump(data, f, indent=2, ensure_ascii=False, default=str)
                
                QMessageBox.information(self, "Başarılı", "Yer imleri başarıyla dışa aktarıldı.")
            except Exception as e:
                QMessageBox.critical(self, "Hata", f"Dışa aktarma hatası: {e}")
    
    def import_bookmarks(self):
        """Yer imlerini içe aktar"""
        file_path, _ = QFileDialog.getOpenFileName(
            self,
            "Yer İmlerini İçe Aktar",
            "",
            "JSON Dosyaları (*.json);;Tüm Dosyalar (*.*)"
        )
        
        if file_path:
            try:
                with open(file_path, 'r', encoding='utf-8') as f:
                    data = json.load(f)
                
                bookmarks = data.get('bookmarks', [])
                imported_count = 0
                
                for bookmark in bookmarks:
                    title = bookmark.get('title', '')
                    url = bookmark.get('url', '')
                    folder = bookmark.get('folder', 'İçe Aktarılan')
                    
                    if title and url:
                        if self.bookmark_manager.add_bookmark(title, url, folder):
                            imported_count += 1
                
                self.load_bookmarks()
                QMessageBox.information(self, "Başarılı", f"{imported_count} yer imi başarıyla içe aktarıldı.")
                
            except Exception as e:
                QMessageBox.critical(self, "Hata", f"İçe aktarma hatası: {e}")

class BookmarkEditDialog(QDialog):
    """
    Yer imi düzenleme dialog'u
    """
    
    def __init__(self, bookmark_manager: BookmarkManager, bookmark: Dict[str, Any] = None, parent=None):
        super().__init__(parent)
        self.bookmark_manager = bookmark_manager
        self.bookmark = bookmark
        self.is_edit = bookmark is not None
        
        self.setWindowTitle("Yer İmi Düzenle" if self.is_edit else "Yeni Yer İmi")
        self.setModal(True)
        self.resize(500, 300)
        
        self.init_ui()
        
        if self.is_edit:
            self.load_bookmark_data()
    
    def init_ui(self):
        """UI'yı başlat"""
        layout = QVBoxLayout(self)
        
        # Form layout
        form_layout = QFormLayout()
        
        # Başlık
        self.title_edit = QLineEdit()
        self.title_edit.setPlaceholderText("Yer imi başlığı...")
        form_layout.addRow("📝 Başlık:", self.title_edit)
        
        # URL
        self.url_edit = QLineEdit()
        self.url_edit.setPlaceholderText("https://example.com")
        form_layout.addRow("🌐 URL:", self.url_edit)
        
        # Klasör
        self.folder_combo = QComboBox()
        self.folder_combo.setEditable(True)
        folders = self.bookmark_manager.get_folders()
        self.folder_combo.addItems(folders)
        if "Genel" not in folders:
            self.folder_combo.addItem("Genel")
        form_layout.addRow("📁 Klasör:", self.folder_combo)
        
        layout.addLayout(form_layout)
        
        # Butonlar
        button_layout = QHBoxLayout()
        button_layout.addStretch()
        
        cancel_btn = QPushButton("❌ İptal")
        cancel_btn.clicked.connect(self.reject)
        button_layout.addWidget(cancel_btn)
        
        save_btn = QPushButton("💾 Kaydet")
        save_btn.clicked.connect(self.save_bookmark)
        save_btn.setDefault(True)
        button_layout.addWidget(save_btn)
        
        layout.addLayout(button_layout)
        
        # Stil
        self.setStyleSheet("""
            QDialog {
                background-color: #f8f9fa;
            }
            QLineEdit, QComboBox {
                border: 2px solid #dee2e6;
                border-radius: 6px;
                padding: 8px;
                font-size: 13px;
            }
            QLineEdit:focus, QComboBox:focus {
                border-color: #007bff;
            }
            QPushButton {
                background-color: #007bff;
                color: white;
                border: none;
                border-radius: 6px;
                padding: 8px 16px;
                font-weight: bold;
            }
            QPushButton:hover {
                background-color: #0056b3;
            }
        """)
    
    def load_bookmark_data(self):
        """Yer imi verilerini yükle"""
        if self.bookmark:
            self.title_edit.setText(self.bookmark.get("title", ""))
            self.url_edit.setText(self.bookmark.get("url", ""))
            
            folder = self.bookmark.get("folder", "Genel")
            index = self.folder_combo.findText(folder)
            if index >= 0:
                self.folder_combo.setCurrentIndex(index)
            else:
                self.folder_combo.setCurrentText(folder)
    
    def save_bookmark(self):
        """Yer imi kaydet"""
        title = self.title_edit.text().strip()
        url = self.url_edit.text().strip()
        folder = self.folder_combo.currentText().strip()
        
        if not title:
            QMessageBox.warning(self, "Uyarı", "Başlık boş olamaz.")
            return
        
        if not url:
            QMessageBox.warning(self, "Uyarı", "URL boş olamaz.")
            return
        
        # URL formatını kontrol et
        if not url.startswith(('http://', 'https://', 'file://')):
            url = 'https://' + url
        
        if not folder:
            folder = "Genel"
        
        try:
            if self.is_edit:
                # Güncelle
                old_url = self.bookmark.get("url", "")
                success = self.bookmark_manager.update_bookmark(old_url, title, url, folder)
            else:
                # Yeni ekle
                success = self.bookmark_manager.add_bookmark(title, url, folder)
            
            if success:
                self.accept()
            else:
                QMessageBox.warning(self, "Uyarı", "Bu URL zaten yer imlerinde mevcut.")
                
        except Exception as e:
            QMessageBox.critical(self, "Hata", f"Kaydetme hatası: {e}") 


"""
Cloud PyIDE - Temel Debug Görünümü
Breakpoint, değişken izleyici, çağrı yığını
"""

import logging
from typing import Dict, List, Optional, Any
from dataclasses import dataclass

try:
    from PyQt6.QtWidgets import *
    from PyQt6.QtCore import *
    from PyQt6.QtGui import *
    PYQT_AVAILABLE = True
except ImportError:
    PYQT_AVAILABLE = False

@dataclass
class DebugBreakpoint:
    """Debug breakpoint"""
    file_path: str
    line_number: int
    enabled: bool = True
    condition: str = ""
    hit_count: int = 0

@dataclass
class DebugVariable:
    """Debug değişkeni"""
    name: str
    value: str
    type_name: str
    scope: str = "local"

@dataclass
class StackFrame:
    """Çağrı yığını frame'i"""
    function_name: str
    file_path: str
    line_number: int
    locals_vars: Dict[str, Any] = None

class DebugManager:
    """Debug yöneticisi"""
    
    def __init__(self, ide_instance=None):
        self.ide_instance = ide_instance
        self.logger = logging.getLogger("DebugManager")
        
        # Debug durumu
        self.is_debugging = False
        self.is_paused = False
        
        # Breakpoint'ler
        self.breakpoints: Dict[str, List[DebugBreakpoint]] = {}
        
        # Debug bilgileri
        self.current_frame: Optional[StackFrame] = None
        self.stack_frames: List[StackFrame] = []
        self.variables: List[DebugVariable] = []
        
        # UI bileşenleri
        self.debug_panel = None
        if PYQT_AVAILABLE:
            self.create_debug_panel()
    
    def create_debug_panel(self) -> QWidget:
        """Debug paneli oluştur"""
        if not PYQT_AVAILABLE:
            return None
        
        # Ana panel
        self.debug_panel = QWidget()
        layout = QVBoxLayout(self.debug_panel)
        
        # Toolbar
        toolbar = QToolBar()
        layout.addWidget(toolbar)
        
        # Debug butonları
        self.start_debug_action = QAction("▶️ Start", self.debug_panel)
        self.start_debug_action.triggered.connect(self.start_debug)
        toolbar.addAction(self.start_debug_action)
        
        self.pause_debug_action = QAction("⏸️ Pause", self.debug_panel)
        self.pause_debug_action.triggered.connect(self.pause_debug)
        self.pause_debug_action.setEnabled(False)
        toolbar.addAction(self.pause_debug_action)
        
        self.stop_debug_action = QAction("⏹️ Stop", self.debug_panel)
        self.stop_debug_action.triggered.connect(self.stop_debug)
        self.stop_debug_action.setEnabled(False)
        toolbar.addAction(self.stop_debug_action)
        
        toolbar.addSeparator()
        
        self.step_over_action = QAction("⏭️ Step Over", self.debug_panel)
        self.step_over_action.triggered.connect(self.step_over)
        self.step_over_action.setEnabled(False)
        toolbar.addAction(self.step_over_action)
        
        self.step_into_action = QAction("⬇️ Step Into", self.debug_panel)
        self.step_into_action.triggered.connect(self.step_into)
        self.step_into_action.setEnabled(False)
        toolbar.addAction(self.step_into_action)
        
        self.step_out_action = QAction("⬆️ Step Out", self.debug_panel)
        self.step_out_action.triggered.connect(self.step_out)
        self.step_out_action.setEnabled(False)
        toolbar.addAction(self.step_out_action)
        
        # Tab widget
        tab_widget = QTabWidget()
        layout.addWidget(tab_widget)
        
        # Breakpoints tab
        self.breakpoints_widget = self.create_breakpoints_widget()
        tab_widget.addTab(self.breakpoints_widget, "🔴 Breakpoints")
        
        # Variables tab
        self.variables_widget = self.create_variables_widget()
        tab_widget.addTab(self.variables_widget, "📊 Variables")
        
        # Call stack tab
        self.stack_widget = self.create_stack_widget()
        tab_widget.addTab(self.stack_widget, "📚 Call Stack")
        
        return self.debug_panel
    
    def create_breakpoints_widget(self) -> QWidget:
        """Breakpoint widget'ı oluştur"""
        widget = QWidget()
        layout = QVBoxLayout(widget)
        
        # Toolbar
        toolbar = QToolBar()
        layout.addWidget(toolbar)
        
        clear_all_action = QAction("🗑️ Clear All", widget)
        clear_all_action.triggered.connect(self.clear_all_breakpoints)
        toolbar.addAction(clear_all_action)
        
        toggle_all_action = QAction("🔄 Toggle All", widget)
        toggle_all_action.triggered.connect(self.toggle_all_breakpoints)
        toolbar.addAction(toggle_all_action)
        
        # Breakpoint listesi
        self.breakpoints_list = QTreeWidget()
        self.breakpoints_list.setHeaderLabels(["File", "Line", "Condition", "Enabled"])
        self.breakpoints_list.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.breakpoints_list.customContextMenuRequested.connect(self.show_breakpoint_context_menu)
        layout.addWidget(self.breakpoints_list)
        
        return widget
    
    def create_variables_widget(self) -> QWidget:
        """Variables widget'ı oluştur"""
        widget = QWidget()
        layout = QVBoxLayout(widget)
        
        # Toolbar
        toolbar = QToolBar()
        layout.addWidget(toolbar)
        
        refresh_action = QAction("🔄 Refresh", widget)
        refresh_action.triggered.connect(self.refresh_variables)
        toolbar.addAction(refresh_action)
        
        # Variables listesi
        self.variables_list = QTreeWidget()
        self.variables_list.setHeaderLabels(["Name", "Value", "Type", "Scope"])
        layout.addWidget(self.variables_list)
        
        return widget
    
    def create_stack_widget(self) -> QWidget:
        """Call stack widget'ı oluştur"""
        widget = QWidget()
        layout = QVBoxLayout(widget)
        
        # Stack listesi
        self.stack_list = QListWidget()
        self.stack_list.itemClicked.connect(self.on_stack_frame_selected)
        layout.addWidget(self.stack_list)
        
        return widget
    
    def add_breakpoint(self, file_path: str, line_number: int, condition: str = "") -> bool:
        """Breakpoint ekle"""
        try:
            if file_path not in self.breakpoints:
                self.breakpoints[file_path] = []
            
            # Aynı satırda breakpoint var mı kontrol et
            for bp in self.breakpoints[file_path]:
                if bp.line_number == line_number:
                    self.logger.warning(f"Breakpoint already exists at {file_path}:{line_number}")
                    return False
            
            # Yeni breakpoint oluştur
            breakpoint = DebugBreakpoint(
                file_path=file_path,
                line_number=line_number,
                condition=condition
            )
            
            self.breakpoints[file_path].append(breakpoint)
            
            # UI'yi güncelle
            self.update_breakpoints_ui()
            
            self.logger.info(f"Added breakpoint at {file_path}:{line_number}")
            return True
            
        except Exception as e:
            self.logger.error(f"Error adding breakpoint: {e}")
            return False
    
    def remove_breakpoint(self, file_path: str, line_number: int) -> bool:
        """Breakpoint kaldır"""
        try:
            if file_path not in self.breakpoints:
                return False
            
            # Breakpoint'i bul ve kaldır
            for i, bp in enumerate(self.breakpoints[file_path]):
                if bp.line_number == line_number:
                    del self.breakpoints[file_path][i]
                    
                    # Liste boşsa dosyayı kaldır
                    if not self.breakpoints[file_path]:
                        del self.breakpoints[file_path]
                    
                    # UI'yi güncelle
                    self.update_breakpoints_ui()
                    
                    self.logger.info(f"Removed breakpoint at {file_path}:{line_number}")
                    return True
            
            return False
            
        except Exception as e:
            self.logger.error(f"Error removing breakpoint: {e}")
            return False
    
    def toggle_breakpoint(self, file_path: str, line_number: int) -> bool:
        """Breakpoint toggle"""
        try:
            if file_path in self.breakpoints:
                for bp in self.breakpoints[file_path]:
                    if bp.line_number == line_number:
                        bp.enabled = not bp.enabled
                        self.update_breakpoints_ui()
                        return True
            
            # Breakpoint yoksa ekle
            return self.add_breakpoint(file_path, line_number)
            
        except Exception as e:
            self.logger.error(f"Error toggling breakpoint: {e}")
            return False
    
    def clear_all_breakpoints(self):
        """Tüm breakpoint'leri temizle"""
        self.breakpoints.clear()
        self.update_breakpoints_ui()
        self.logger.info("Cleared all breakpoints")
    
    def toggle_all_breakpoints(self):
        """Tüm breakpoint'leri toggle et"""
        for file_breakpoints in self.breakpoints.values():
            for bp in file_breakpoints:
                bp.enabled = not bp.enabled
        
        self.update_breakpoints_ui()
    
    def update_breakpoints_ui(self):
        """Breakpoint UI'sini güncelle"""
        if not PYQT_AVAILABLE or not self.breakpoints_list:
            return
        
        self.breakpoints_list.clear()
        
        for file_path, file_breakpoints in self.breakpoints.items():
            for bp in file_breakpoints:
                item = QTreeWidgetItem()
                item.setText(0, file_path)
                item.setText(1, str(bp.line_number))
                item.setText(2, bp.condition)
                item.setText(3, "✓" if bp.enabled else "✗")
                
                # Renk kodlama
                if bp.enabled:
                    item.setForeground(0, QColor("#ff0000"))
                else:
                    item.setForeground(0, QColor("#808080"))
                
                self.breakpoints_list.addTopLevelItem(item)
    
    def start_debug(self):
        """Debug başlat"""
        try:
            self.is_debugging = True
            self.is_paused = False
            
            # UI durumunu güncelle
            self.update_debug_ui_state()
            
            self.logger.info("Debug session started")
            
            # IDE'ye bildir
            if self.ide_instance and hasattr(self.ide_instance, 'on_debug_started'):
                self.ide_instance.on_debug_started()
                
        except Exception as e:
            self.logger.error(f"Error starting debug: {e}")
    
    def pause_debug(self):
        """Debug duraklat"""
        self.is_paused = True
        self.update_debug_ui_state()
        self.logger.info("Debug session paused")
    
    def stop_debug(self):
        """Debug durdur"""
        try:
            self.is_debugging = False
            self.is_paused = False
            
            # Debug bilgilerini temizle
            self.current_frame = None
            self.stack_frames.clear()
            self.variables.clear()
            
            # UI'yi güncelle
            self.update_debug_ui_state()
            self.update_variables_ui()
            self.update_stack_ui()
            
            self.logger.info("Debug session stopped")
            
            # IDE'ye bildir
            if self.ide_instance and hasattr(self.ide_instance, 'on_debug_stopped'):
                self.ide_instance.on_debug_stopped()
                
        except Exception as e:
            self.logger.error(f"Error stopping debug: {e}")
    
    def step_over(self):
        """Step over"""
        self.logger.info("Step over")
        # TODO: Gerçek debug implementasyonu
    
    def step_into(self):
        """Step into"""
        self.logger.info("Step into")
        # TODO: Gerçek debug implementasyonu
    
    def step_out(self):
        """Step out"""
        self.logger.info("Step out")
        # TODO: Gerçek debug implementasyonu
    
    def update_debug_ui_state(self):
        """Debug UI durumunu güncelle"""
        if not PYQT_AVAILABLE:
            return
        
        # Buton durumları
        if hasattr(self, 'start_debug_action'):
            self.start_debug_action.setEnabled(not self.is_debugging)
            self.pause_debug_action.setEnabled(self.is_debugging and not self.is_paused)
            self.stop_debug_action.setEnabled(self.is_debugging)
            
            self.step_over_action.setEnabled(self.is_debugging and self.is_paused)
            self.step_into_action.setEnabled(self.is_debugging and self.is_paused)
            self.step_out_action.setEnabled(self.is_debugging and self.is_paused)
    
    def refresh_variables(self):
        """Değişkenleri yenile"""
        # TODO: Gerçek debug implementasyonu
        self.update_variables_ui()
    
    def update_variables_ui(self):
        """Variables UI'sini güncelle"""
        if not PYQT_AVAILABLE or not self.variables_list:
            return
        
        self.variables_list.clear()
        
        for var in self.variables:
            item = QTreeWidgetItem()
            item.setText(0, var.name)
            item.setText(1, var.value)
            item.setText(2, var.type_name)
            item.setText(3, var.scope)
            
            self.variables_list.addTopLevelItem(item)
    
    def update_stack_ui(self):
        """Call stack UI'sini güncelle"""
        if not PYQT_AVAILABLE or not self.stack_list:
            return
        
        self.stack_list.clear()
        
        for i, frame in enumerate(self.stack_frames):
            text = f"{frame.function_name} - {frame.file_path}:{frame.line_number}"
            item = QListWidgetItem(text)
            item.setData(Qt.ItemDataRole.UserRole, i)
            self.stack_list.addItem(item)
    
    def on_stack_frame_selected(self, item):
        """Stack frame seçildi"""
        frame_index = item.data(Qt.ItemDataRole.UserRole)
        if 0 <= frame_index < len(self.stack_frames):
            self.current_frame = self.stack_frames[frame_index]
            self.logger.info(f"Selected stack frame: {self.current_frame.function_name}")
    
    def show_breakpoint_context_menu(self, position):
        """Breakpoint context menü"""
        if not PYQT_AVAILABLE:
            return
        
        item = self.breakpoints_list.itemAt(position)
        if not item:
            return
        
        menu = QMenu()
        
        edit_action = menu.addAction("Edit Condition")
        edit_action.triggered.connect(lambda: self.edit_breakpoint_condition(item))
        
        toggle_action = menu.addAction("Toggle")
        toggle_action.triggered.connect(lambda: self.toggle_breakpoint_from_ui(item))
        
        delete_action = menu.addAction("Delete")
        delete_action.triggered.connect(lambda: self.delete_breakpoint_from_ui(item))
        
        menu.exec(self.breakpoints_list.mapToGlobal(position))
    
    def edit_breakpoint_condition(self, item):
        """Breakpoint koşulunu düzenle"""
        # TODO: Koşul düzenleme dialog'u
        pass
    
    def toggle_breakpoint_from_ui(self, item):
        """UI'den breakpoint toggle"""
        file_path = item.text(0)
        line_number = int(item.text(1))
        self.toggle_breakpoint(file_path, line_number)
    
    def delete_breakpoint_from_ui(self, item):
        """UI'den breakpoint sil"""
        file_path = item.text(0)
        line_number = int(item.text(1))
        self.remove_breakpoint(file_path, line_number)
    
    def get_debug_panel(self) -> Optional[QWidget]:
        """Debug panelini al"""
        return self.debug_panel
    
    def has_breakpoint(self, file_path: str, line_number: int) -> bool:
        """Breakpoint var mı?"""
        if file_path not in self.breakpoints:
            return False
        
        for bp in self.breakpoints[file_path]:
            if bp.line_number == line_number:
                return True
        
        return False
    
    def get_breakpoints_for_file(self, file_path: str) -> List[DebugBreakpoint]:
        """Dosya için breakpoint'leri al"""
        return self.breakpoints.get(file_path, []) 


"""
Core Thread Pool - Thread Havuzu Yöneticisi  
Thread yoğunluğunu sınırlı sayıda thread ile dengeleyen sistemsel yürütme havuzu
"""

import logging
import threading
import time
import concurrent.futures
import os
from typing import Dict, List, Optional, Any, Callable
from datetime import datetime
from dataclasses import dataclass
import uuid
import psutil

# Python 3.8+ için Future import
try:
    from typing import Future
except ImportError:
    from concurrent.futures import Future

@dataclass
class PoolStats:
    """Thread pool istatistikleri"""
    active_threads: int = 0
    idle_threads: int = 0
    completed_tasks: int = 0
    failed_tasks: int = 0
    average_task_time: float = 0.0
    peak_threads: int = 0
    total_uptime: float = 0.0

class ThreadPool:
    """
    Thread yoğunluğunu sınırlı sayıda thread ile dengeleyen sistemsel yürütme havuzu.
    Çekirdek sayısına göre sabit havuz oluşturma, aşırı yoğunlukta bekletme.
    """
    
    def __init__(self, kernel, max_workers: Optional[int] = None):
        self.kernel = kernel
        self.logger = logging.getLogger("ThreadPool")
        
        # Worker sayısını belirle
        if max_workers is None:
            # CPU çekirdek sayısının 2 katı (I/O intensive işler için)
            max_workers = min(32, (os.cpu_count() or 1) * 2)
        
        self.max_workers = max_workers
        self.min_workers = max(1, max_workers // 4)  # Minimum %25
        
        # Thread pool executor
        self.executor: Optional[concurrent.futures.ThreadPoolExecutor] = None
        self.running = False
        self.start_time = None
        
        # Görev takibi
        self.active_futures: Dict[str, Future] = {}
        self.task_results: Dict[str, Any] = {}
        
        # İstatistikler
        self.stats = PoolStats()
        self.task_completion_times: List[float] = []
        
        # Monitoring
        self.monitor_thread: Optional[threading.Thread] = None
        self.monitor_active = False
        
        # Thread limit ve backpressure
        self.task_limit = max_workers * 10  # Maksimum bekleyen görev sayısı
        self.backpressure_enabled = True
        
        # Locks
        self.stats_lock = threading.RLock()
        
    def start(self):
        """Thread pool'u başlat"""
        try:
            if self.running:
                return
            
            # ThreadPoolExecutor'ı başlat
            self.executor = concurrent.futures.ThreadPoolExecutor(
                max_workers=self.max_workers,
                thread_name_prefix="PyCloudPool"
            )
            
            self.running = True
            self.start_time = time.time()
            
            # Monitoring başlat
            self._start_monitoring()
            
            self.logger.info(f"Thread pool started with {self.max_workers} max workers")
            
        except Exception as e:
            self.logger.error(f"Failed to start thread pool: {e}")
    
    def stop(self, wait: bool = True, timeout: Optional[float] = None):
        """Thread pool'u durdur"""
        try:
            if not self.running:
                return
            
            self.running = False
            
            # Monitoring'i durdur
            self._stop_monitoring()
            
            # Executor'ı kapat
            if self.executor:
                self.executor.shutdown(wait=wait, timeout=timeout)
                self.executor = None
            
            # İstatistikleri güncelle
            if self.start_time:
                self.stats.total_uptime = time.time() - self.start_time
            
            self.logger.info("Thread pool stopped")
            
        except Exception as e:
            self.logger.error(f"Failed to stop thread pool: {e}")
    
    def submit(self, function: Callable, *args, **kwargs) -> Optional[str]:
        """Thread pool'a görev gönder"""
        try:
            if not self.running or not self.executor:
                self.logger.warning("Thread pool not running")
                return None
            
            # Backpressure kontrolü
            if self.backpressure_enabled and len(self.active_futures) >= self.task_limit:
                self.logger.warning("Thread pool at capacity, rejecting task")
                return None
            
            # Görevi gönder
            task_id = str(uuid.uuid4())
            future = self.executor.submit(self._execute_with_tracking, function, task_id, *args, **kwargs)
            
            # Takip listesine ekle
            self.active_futures[task_id] = future
            
            self.logger.debug(f"Task submitted: {task_id}")
            return task_id
            
        except Exception as e:
            self.logger.error(f"Failed to submit task: {e}")
            return None
    
    def submit_batch(self, tasks: List[tuple]) -> List[str]:
        """Toplu görev gönderimi"""
        try:
            task_ids = []
            
            for task in tasks:
                if len(task) >= 1:
                    function = task[0]
                    args = task[1:] if len(task) > 1 else ()
                    task_id = self.submit(function, *args)
                    if task_id:
                        task_ids.append(task_id)
            
            self.logger.info(f"Batch submitted: {len(task_ids)} tasks")
            return task_ids
            
        except Exception as e:
            self.logger.error(f"Failed to submit batch: {e}")
            return []
    
    def get_result(self, task_id: str, timeout: Optional[float] = None) -> Any:
        """Görev sonucunu al"""
        try:
            # Önce cache'den kontrol et
            if task_id in self.task_results:
                return self.task_results[task_id]
            
            # Future'dan al
            if task_id in self.active_futures:
                future = self.active_futures[task_id]
                result = future.result(timeout=timeout)
                
                # Cache'e ekle
                self.task_results[task_id] = result
                
                return result
            
            return None
            
        except concurrent.futures.TimeoutError:
            self.logger.warning(f"Task {task_id} timed out")
            return None
        except Exception as e:
            self.logger.error(f"Failed to get result for task {task_id}: {e}")
            return None
    
    def wait_for_completion(self, task_ids: List[str], timeout: Optional[float] = None) -> Dict[str, Any]:
        """Birden fazla görevin tamamlanmasını bekle"""
        try:
            results = {}
            futures_to_wait = {}
            
            # Future'ları topla
            for task_id in task_ids:
                if task_id in self.active_futures:
                    futures_to_wait[self.active_futures[task_id]] = task_id
            
            # Tamamlanmasını bekle
            completed_futures = concurrent.futures.as_completed(futures_to_wait.keys(), timeout=timeout)
            
            for future in completed_futures:
                task_id = futures_to_wait[future]
                try:
                    result = future.result()
                    results[task_id] = result
                    self.task_results[task_id] = result
                except Exception as e:
                    results[task_id] = {"error": str(e)}
                    self.logger.error(f"Task {task_id} failed: {e}")
            
            return results
            
        except concurrent.futures.TimeoutError:
            self.logger.warning("Wait for completion timed out")
            return {}
        except Exception as e:
            self.logger.error(f"Failed to wait for completion: {e}")
            return {}
    
    def _execute_with_tracking(self, function: Callable, task_id: str, *args, **kwargs):
        """Takip ile görev çalıştır"""
        start_time = time.time()
        
        try:
            # Aktif thread sayısını artır
            with self.stats_lock:
                self.stats.active_threads += 1
                if self.stats.active_threads > self.stats.peak_threads:
                    self.stats.peak_threads = self.stats.active_threads
            
            # Fonksiyonu çalıştır
            result = function(*args, **kwargs)
            
            # Başarılı tamamlama
            execution_time = time.time() - start_time
            
            with self.stats_lock:
                self.stats.completed_tasks += 1
                self.task_completion_times.append(execution_time)
                self._update_average_task_time()
            
            return result
            
        except Exception as e:
            # Hata durumu
            with self.stats_lock:
                self.stats.failed_tasks += 1
            
            self.logger.error(f"Task {task_id} failed: {e}")
            raise
            
        finally:
            # Aktif thread sayısını azalt
            with self.stats_lock:
                self.stats.active_threads = max(0, self.stats.active_threads - 1)
            
            # Future'ı temizle
            if task_id in self.active_futures:
                del self.active_futures[task_id]
    
    def _update_average_task_time(self):
        """Ortalama görev süresini güncelle"""
        try:
            if self.task_completion_times:
                # Son 1000 görevin ortalamasını al
                recent_times = self.task_completion_times[-1000:]
                self.stats.average_task_time = sum(recent_times) / len(recent_times)
                
        except Exception as e:
            self.logger.debug(f"Failed to update average task time: {e}")
    
    def _start_monitoring(self):
        """Pool monitoring'ini başlat"""
        try:
            self.monitor_active = True
            self.monitor_thread = threading.Thread(
                target=self._monitor_loop,
                name="ThreadPoolMonitor",
                daemon=True
            )
            self.monitor_thread.start()
            
        except Exception as e:
            self.logger.error(f"Failed to start monitoring: {e}")
    
    def _stop_monitoring(self):
        """Pool monitoring'ini durdur"""
        try:
            self.monitor_active = False
            if self.monitor_thread:
                self.monitor_thread.join(timeout=5)
                
        except Exception as e:
            self.logger.error(f"Failed to stop monitoring: {e}")
    
    def _monitor_loop(self):
        """Monitoring ana döngüsü"""
        while self.monitor_active and self.running:
            try:
                self._update_pool_stats()
                self._cleanup_completed_tasks()
                time.sleep(10)  # 10 saniyede bir kontrol
                
            except Exception as e:
                self.logger.error(f"Monitor loop error: {e}")
                time.sleep(5)
    
    def _update_pool_stats(self):
        """Pool istatistiklerini güncelle"""
        try:
            with self.stats_lock:
                # İdeal thread sayısı
                total_threads = len(threading.enumerate())
                self.stats.idle_threads = max(0, self.max_workers - self.stats.active_threads)
                
            # Sistem kaynak kullanımı
            if hasattr(psutil, 'cpu_percent'):
                cpu_usage = psutil.cpu_percent(interval=0.1)
                if cpu_usage > 80:  # CPU yoğunluğu kontrolü
                    self.logger.debug(f"High CPU usage detected: {cpu_usage}%")
                    
        except Exception as e:
            self.logger.debug(f"Failed to update pool stats: {e}")
    
    def _cleanup_completed_tasks(self):
        """Tamamlanmış görevleri temizle"""
        try:
            # Eski task sonuçlarını temizle (1000'den fazla varsa)
            if len(self.task_results) > 1000:
                # En eskilerini sil
                items = list(self.task_results.items())
                self.task_results = dict(items[-500:])  # Son 500'ü koru
                
            # Tamamlanmış future'ları temizle
            completed_tasks = []
            for task_id, future in self.active_futures.items():
                if future.done():
                    completed_tasks.append(task_id)
            
            for task_id in completed_tasks:
                del self.active_futures[task_id]
                
        except Exception as e:
            self.logger.debug(f"Failed to cleanup completed tasks: {e}")
    
    def get_stats(self) -> Dict[str, Any]:
        """Pool istatistiklerini al"""
        with self.stats_lock:
            uptime = time.time() - self.start_time if self.start_time else 0
            
            return {
                "max_workers": self.max_workers,
                "active_threads": self.stats.active_threads,
                "idle_threads": self.stats.idle_threads,
                "completed_tasks": self.stats.completed_tasks,
                "failed_tasks": self.stats.failed_tasks,
                "average_task_time": self.stats.average_task_time,
                "peak_threads": self.stats.peak_threads,
                "pending_tasks": len(self.active_futures),
                "cached_results": len(self.task_results),
                "uptime_seconds": uptime,
                "running": self.running
            }
    
    def resize_pool(self, new_max_workers: int):
        """Pool boyutunu değiştir (dinamik)"""
        try:
            if not self.running:
                self.max_workers = new_max_workers
                return
            
            # Yeni executor oluştur
            old_executor = self.executor
            
            self.executor = concurrent.futures.ThreadPoolExecutor(
                max_workers=new_max_workers,
                thread_name_prefix="PyCloudPool"
            )
            
            self.max_workers = new_max_workers
            
            # Eski executor'ı kapat
            if old_executor:
                old_executor.shutdown(wait=False)
            
            self.logger.info(f"Thread pool resized to {new_max_workers} workers")
            
        except Exception as e:
            self.logger.error(f"Failed to resize pool: {e}")
    
    def force_cleanup(self):
        """Zorla temizlik yap"""
        try:
            # Tüm tamamlanmış task'ları temizle
            self.task_results.clear()
            
            # Tamamlanmış future'ları temizle
            completed_tasks = []
            for task_id, future in self.active_futures.items():
                if future.done():
                    completed_tasks.append(task_id)
            
            for task_id in completed_tasks:
                del self.active_futures[task_id]
            
            # İstatistikleri sıfırla
            with self.stats_lock:
                self.task_completion_times = self.task_completion_times[-100:]  # Son 100'ü koru
            
            self.logger.info("Force cleanup completed")
            
        except Exception as e:
            self.logger.error(f"Failed to force cleanup: {e}") 